/*!
 *  @file EEPROM_settings.h
 *
 */

#ifndef _EEPROM_SETTINGS_H
#define _EEPROM_SETTINGS_H

#define N_MagneticDensitySensors 5

typedef struct
{
    int ISM303_INT_acc_x_minus100pcReading;
    int ISM303_INT_acc_x_plus100pcReading;
    int ISM303_INT_acc_y_minus100pcReading;
    int ISM303_INT_acc_y_plus100pcReading;
    int ISM303_INT_acc_z_minus100pcReading;
    int ISM303_INT_acc_z_plus100pcReading;

    int ISM303_INT_mag_x_minus100pcReading;
    int ISM303_INT_mag_x_plus100pcReading;
    int ISM303_INT_mag_y_minus100pcReading;
    int ISM303_INT_mag_y_plus100pcReading;
    int ISM303_INT_mag_z_minus100pcReading;
    int ISM303_INT_mag_z_plus100pcReading;

    int MagneticSensors0pcReadings[N_MagneticDensitySensors];
    int MagneticSensors100pcReadings[N_MagneticDensitySensors];

    int ControlMode;
    int RPMsetFast;
    int RPMsetNormal;
    int TautoTrigger;
    int TFoffset;
    int ParkOrientation;
    int Vmin;
    int Vmax;
    int d_MB_MC;
    int TautoTurn;
    int BitFlags_0;
    int d_MD_ME;
    int V_ratio;
    int spare1;
    int spare2;
    int spare3;
    int spare4;
  
    uint32_t crc32value;
    
} SettingsBuffer_t;


#define SettingsBufferSize sizeof(SettingsBuffer_t)
#define Nbackups                        3
#define EEPROM_Settings_start_address   (256)
#define EEPROM_Settings_end_address     (EEPROM_Settings_start_address + Nbackups*SettingsBufferSize)

class EEPROM_settings 
{
public:
    EEPROM_settings();

    void Default_EEPROM_Settings (int mode);
    void Write_EEPROM_Settings ();
    bool Read_EEPROM_Settings ();
    void show_all_EEPROM_settings_bytes ();

    SettingsBuffer_t SettingsBuffer;

private:
    bool Load_EEPROM_Settings (int WhichBuffer);
    void Save_EEPROM_Settings (int WhichBuffer);
    void AddCRC32 ();
    bool CheckCRC32 ();
    void show_EEPROM_settings_bytes (int WhichBuffer);
};

#endif
