/*!
 *  @file Enc50t_Protocol.cpp
 *
 * 	BSD (see license.txt)
 *
 */

#include "Enc50t_Protocol.h"

#include <TimeLib.h>
#include <util/atomic.h>


int g_encoder_type = 0;

uint16_t Encoder50t = 0;

uint16_t EncWordToPC = 0;

uint16_t ENCnibbleCounter = 0;
uint16_t ENCindex = 0; 
uint16_t ENCnibble = 0;
uint32_t ENCdword = 0;
uint32_t ENCdword2 = 0;

#define ENCmessageMaxLength 10
uint32_t ENCmessage [ENCmessageMaxLength];

uint8_t EncAsciiToHexNibble(char c)
{
  if (c<'0')
      return 0;
  if (c>'F')
      c -= 32;  // toupper

  if (c>'F')
      return 0;
  if ((c>'9') && (c<'A'))
      return 0;

  c -= '0';
  if (c > 9)
      c -= 7;

  return (c & 0xf);
}

void ParseEnc50t_character (char ch)
{
    uint16_t NextEncWordToPC;
    if (ch == 10)
    {
        if (ENCindex == 2)
        {
            if ((ENCmessage[0] ^ ENCmessage[1]) == 0xffffff) 
              switch (g_encoder_type)
              {
                case 0: // A,B,Z
                    Encoder50t = (ENCmessage[0] & 0xfff) << 2; // assuming 12 bits are sent, so 0..4095 for 0-360 deg - change that to 14 bits
                    EncWordToPC = 0x4000 | Encoder50t;
                    break;
                    
                case 1: // SSI
                    if (ENCmessage[0] & 0x800000)   // Position valid bit (D23 Celera) is set?
                        Encoder50t = (ENCmessage[0] >> 4) & 0x3fff; // assuming it's 18 bits for 0-360 deg, reduce that here to 14 bits
                        
                    NextEncWordToPC = (ENCmessage[0] >> 4) & 0x3fff; // bits[13:0] for 0=360 deg, b15 at zero, b14 = Position valid bit
                    if (ENCmessage[0] & 0x800000)
                        NextEncWordToPC |= 0x4000;
                    EncWordToPC = NextEncWordToPC;    
                    break;
              }
        }
    }
    else
        if (((ch >= '0') && (ch <= '9')) || 
            ((ch >= 'A') && (ch <= 'F')) || 
            ((ch >= 'a') && (ch <= 'f')))
        {
            ENCnibble = EncAsciiToHexNibble(ch);
            switch (ENCnibbleCounter)
            {
                case 0: ENCdword = ENCnibble;
                        ENCnibbleCounter++;
                        break;
                case 1:
                case 2:
                case 3:
                case 4: ENCdword <<= 4;
                        ENCdword += ENCnibble;
                        ENCnibbleCounter++;
                        break;
                case 5: ENCdword <<= 4;
                        ENCdword += ENCnibble;
                        ENCmessage[ENCindex++] = ENCdword;
                        if (ENCindex >= ENCmessageMaxLength)
                            ENCindex--;
                        ENCnibbleCounter++;
                        break;

                case 6: ENCdword2 = ENCnibble;
                        ENCnibbleCounter++;
                        break;
                case 7:
                case 8:
                case 9:
                case 10: ENCdword2 <<= 4;
                        ENCdword2 += ENCnibble;
                        ENCnibbleCounter++;
                        break;
                case 11: ENCdword2 <<= 4;
                        ENCdword2 += ENCnibble;
                        ENCmessage[ENCindex++] = ENCdword2;
                        if (ENCindex >= ENCmessageMaxLength)
                            ENCindex--;
                        ENCnibbleCounter = 0;
            }
        }
        else
        {
            ENCindex = 0;
            ENCnibbleCounter= 0;
        }
}

void serialEvent2() 
{
    while (Serial2.available())
    {
        char ch = Serial2.read();
        {
//            Serial.print (ch);
            ParseEnc50t_character (ch);
        }
    }
}

void send_poll_for_50t_encoder (int SSI_encoder)
{   
    g_encoder_type = SSI_encoder;
    
    if (SSI_encoder)
        Serial2.print ('$');
    else
        Serial2.print ('?');
 }
