/*!
 *  @file HostProtocol.cpp
 *
 * 	BSD (see license.txt)
 *
 */

#include "HostProtocol.h"
#include "EEPROM_settings.h"

#include <TimeLib.h>
#include <util/atomic.h>

uint16_t our_address = 0x2;

extern volatile char versionString[];

extern EEPROM_settings OurSettings;

extern float exit_shaft_orientation, measured_diverter_orientation, setpoint_diverter_orientation;
extern float RPM_exit_shaft, setpoint_RPM_exit_shaft;
extern float MA, MB, MC, MD, ME;
extern float motor_current;
extern uint16_t ShaftEncoderAndPremexReedSwitch;

extern int ProcessPoll_ISM303_INT (int comm, int param, int16_t *resp);


char HOSTresp[HOSTmessageMaxLength*4+10];
uint16_t HOSTresp_length;

int haveReply = 0;
int haveReplyButWaitTillRS485Free = 0;

int ri, rj, rk, rm, rn, rt;
int new_ijkmnt = 0; 

int16_t Xcorr_Velocity_as_int16 = 0;
int16_t Xcorr_Corr_Coeff_as_int16 = 0;
int16_t Spare3 = 0;
int16_t FIFO_delay_D_to_Bit = 0;

volatile uint8_t HOSTnibble;
volatile uint16_t HOSTword;
volatile uint16_t HOSTmessage[HOSTmessageMaxLength+2];
volatile uint16_t HOSTindex=0;
volatile uint16_t HOSTnibbleCounter;
volatile uint16_t CRC16w;

#define N_10ms_VALUES 11
#define ELASTIC_BUFFER_SIZE 100
#define ELASTIC_BUFFER_NWORDS (N_10ms_VALUES * ELASTIC_BUFFER_SIZE)

uint16_t elastic_buffer[ELASTIC_BUFFER_NWORDS];
uint32_t BufferWritePointer = 0;
uint32_t BufferReadPointer = 0;
int BufferUsed = 0;

char prev_ch = 0;
bool ReadableProtocol = false;
bool DLEDLE_history = false;
#define STX 0x02
#define ETX 0x03
#define DLE 0x10

int32_t Cnt = 0;

void serialEvent1() 
{
    while (Serial1.available())
    {
        char ch = Serial1.read();
        {
 //           Serial.print (ch);
            ParseHOSTcharacter (ch);
        }
    }
}

uint16_t int16inuint16 (int16_t x)
{
    return (uint16_t)x;
}  

void push_one_sample (uint16_t *M)
{
    *M++ = int16inuint16 ((int16_t)(16384.0 * measured_diverter_orientation / 360.0));
    *M++ = int16inuint16 ((int16_t)(16384.0 * setpoint_diverter_orientation / 360.0));
                    
    *M++ = int16inuint16 ((int16_t)(16384.0 * RPM_exit_shaft / NREF));
    *M++ = int16inuint16 ((int16_t)(16384.0 * setpoint_RPM_exit_shaft / NREF));
                    
    *M++ = int16inuint16 ((int16_t)(16384.0 * motor_current / IREF));
                  
    *M++ = int16inuint16 ((int16_t)(16384.0 * MA / MREF));
    *M++ = int16inuint16 ((int16_t)(16384.0 * MB / MREF));
 //*M++ = 0;
 //*M++ = int16inuint16 ((int16_t)(Cnt & 0x3fff));
    Cnt += 100;
    
    *M++ = int16inuint16 ((int16_t)(16384.0 * MC / MREF));
    *M++ = int16inuint16 ((int16_t)(16384.0 * MD / MREF));

    *M++ = int16inuint16 ((int16_t)(16384.0 * ME / MREF));

    *M++ = ShaftEncoderAndPremexReedSwitch;
    
}

uint32_t nextPointer (uint32_t p)
{
    p += N_10ms_VALUES;
    if (p >= ELASTIC_BUFFER_NWORDS)
        p = 0;         
    return p;
}

void push_10ms_values_in_elastic_buffer ()
{
    push_one_sample (&elastic_buffer[BufferWritePointer]);
    BufferWritePointer = nextPointer (BufferWritePointer);
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE) 
    {
        if (BufferUsed < ELASTIC_BUFFER_SIZE)
            BufferUsed++;
        else
        {
            BufferReadPointer = nextPointer (BufferReadPointer); // bin what no longer fits in buffer
//Serial.printf (" BIN ");
        }
    }
}

int pop_10ms_values_from_elastic_buffer (uint16_t *p)
{
    int i;
    int r = 0;
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE) 
    {
        if (BufferUsed)
        {
            for (i=0; i<N_10ms_VALUES; i++)
                *p++ = elastic_buffer[BufferReadPointer + i];
            BufferReadPointer = nextPointer (BufferReadPointer);
            r = BufferUsed;
            BufferUsed--;
        }
    }
    return r;   
}

void GetDateTime(uint16_t *wp) 
{
    TimeElements tm;
      
    time_t timenow = Teensy3Clock.get();
    breakTime(timenow, tm);
    
    wp[2] = tm.Day;
    wp[1] = tm.Month;
    wp[0] = 1970+tm.Year;
    
    wp[3] = tm.Hour;
    wp[4] = tm.Minute;
    wp[5] = tm.Second;
}

void SetDateTime (uint16_t *t)
{
    TimeElements tm;
    time_t t_set;
    
    tm.Year = 2000 - 1970 + t[0];
    tm.Month = t[1];
    tm.Day = t[2];
    tm.Hour = t[3];
    tm.Minute = t[4];
    tm.Second = t[5];
    
    t_set = makeTime(tm);

    Teensy3Clock.set(t_set);
}

void Add_to_CRC16 (uint8_t newx)
{
    uint8_t x;
    x = (CRC16w >> 8) ^ newx;
    x ^= x>>4;
    CRC16w = (CRC16w << 8) ^ ((uint16_t)(x << 12)) ^ ((uint16_t)(x <<5)) ^ ((uint16_t)x);
}

void Add_to_CRC16w (uint16_t w)
{
   Add_to_CRC16 (w & 0xff);
   Add_to_CRC16 (w >> 8);
}

uint16_t Calc_Crc16 (uint8_t *bp, uint16_t Nbytes)
{
    CRC16w = 0xffff;
    while (Nbytes--)
        Add_to_CRC16(*bp++);
    return CRC16w;
}

void putword (char *s, uint16_t x)
{
   sprintf (s, "%04X", x);
}

int putbyte_mind_DLEs (uint8_t *s, uint8_t x)
{
    int i;
    if (x == DLE)
    {
        s[0] = DLE;
        s[1] = x;
        i = 2;
    }
    else
    {
        s[0] = x;
        i = 1;
    }      
         
    return i;
}

int HOSTreplyMessage (uint16_t *Message, uint16_t MessageLength, char *s)
{
    if (MessageLength <= HOSTmessageMaxLength)
    {
        Message[1] = MessageLength & 0x7ff;
        Message[2] |= 0x8000;   // response has b15 set in command word

        int i = 0;
        
        s[i++] = 0x01;  // dummy to avoid collission with tail of previous message
        
        if (ReadableProtocol)       
        {
            s[i++] = 'x';
        }
        else
        {
            s[i++] = DLE;
            s[i++] = STX;
        }
        
        CRC16w = 0xffff;

        MessageLength += 3;  // word0 = source/dest, word1 = length, word2 = command/response code
        while (MessageLength--)
        {
            if (ReadableProtocol) 
            {
                putword (&s[i], *Message);
                i += 4;
            }
            else
            {
                i += putbyte_mind_DLEs ((uint8_t*)&s[i], *Message >> 8);
                i += putbyte_mind_DLEs ((uint8_t*)&s[i], *Message & 0xff);
            }
            Add_to_CRC16w (*Message++);
        }

        
        if (ReadableProtocol) 
        {
            putword (&s[i], CRC16w);
            i += 4;
        }
        else
        {
            i += putbyte_mind_DLEs ((uint8_t*)&s[i], CRC16w >> 8);
            i += putbyte_mind_DLEs ((uint8_t*)&s[i], CRC16w & 0xff);
        }
        
        if (ReadableProtocol) 
        {
            s[i++] = 'X';
            s[i++] = 0x0a;    // line feed for dosend trigger
            s[i++] = 0; // end of text string
        }
        else
        {
            s[i++] = DLE;
            s[i++] = ETX;
        }
       
        HOSTresp_length = i;
        return 1;
    }
    else
    {
        HOSTresp_length = 0;
        return 0;
    }
}

int ProcessHOSTmessage (uint16_t *Message, uint16_t MessageLength, char *st)
{
    int r = 0;
    int i = 3;
    int j;
//    uint16_t w;

// first message word is <5 bit source address><5 bits destination address><6 bits length of payload, counted in words following the command word>
// second word is command
// remainding words is optional command parameters, 0..length-1 words

    uint16_t source_address = (Message[0] >> 8) & 0xff;
    uint16_t dest_address = Message[0] & 0xff;
    uint16_t len = Message[1] & 0x7ff;

    if (dest_address == our_address)
    {
        Message[0] = (dest_address << 8) | source_address;  // roles swap
        
        switch (Message[2]) // command
        {
            case 0:                           // test echo (invert testdata)
                for (i=0; i<len; i++)
                    Message[3+i] = Message[3+i] ^ 0xffff;
                    
                r = HOSTreplyMessage (Message, len, st);
                break;
            case 1:                           // get version string
                if (len == 0)
                {
                    j = strlen((const char*)versionString);
                    if (j < (HOSTmessageMaxLength * 2))
                    {
                        int k = 0;
                        for (i=0; i<((j/2)+1); i++)
                        {
                            Message[3+i] = versionString[k++];
                            Message[3+i] += ((uint16_t)versionString[k++])<<8;
                              
                        }
                        r = HOSTreplyMessage (Message, 1 + (j+1)/2, st);
                    }
                }
                break;
            case 2:                           // get timedate
                if (len == 0)
                {
                    GetDateTime (&Message[3]);
                    r = HOSTreplyMessage (Message, 6, st);
                }
                break;
            case 3:                           // set timedate
                if (len == 7)
                {
                    if (Message[i++] == 0xaa55)
                    {
                      SetDateTime (&Message[i]);
                      r = HOSTreplyMessage (Message, 0, st);
                    }
                }
                break;
            case 4:                           // set i j k m n t
                if (len == 6)
                {
                    ri = Message[i++];
                    rj = Message[i++];
                    rk = Message[i++];
                    rm = Message[i++];
                    rn = Message[i++];
                    rt = Message[i++];
                    new_ijkmnt = 1; 
                    r = HOSTreplyMessage (Message, 0, st);
                }
                break;
            case 5:                           // read settings
                if (len == 0)
                {
                    memcpy ((uint8_t*)&Message[i], (uint8_t*)&OurSettings, sizeof(SettingsBuffer_t));
                    r = HOSTreplyMessage (Message, sizeof(SettingsBuffer_t) >> 1, st);
                }
                break;
            case 6:                           // write settings
                if (Message[i++] == 0x55aa)   // key = 0x55aa
                    if (len == 1 + (sizeof(SettingsBuffer_t) >> 1))
                    {
                        memcpy ((uint8_t*)&OurSettings, (uint8_t*)&Message[i], sizeof(SettingsBuffer_t));
                        r = HOSTreplyMessage (Message, 0, st);
                    }
                    
                break;
            case 7:                           // load settings from Teensy EEPROM
                if (len == 1)
                    if (Message[i++] == 0x55ab)   // key = 0x55ab
                    {
                        OurSettings.Read_EEPROM_Settings ();
                        r = HOSTreplyMessage (Message, 0, st);
                    }

                break;
            case 8:                           // save settings in Teesny EEPROM
                if (len == 1)
                    if (Message[i++] == 0x55ac)   // key = 0x55ac
                    {
                        OurSettings.Write_EEPROM_Settings ();
                        r = HOSTreplyMessage (Message, 0, st);
                    }
                break;
            case 9:                           // save default settings in Teesny EEPROM
                if (len == 1)
                    if (Message[i++] == 0x55ad)   // key = 0x55ad
                    {
                        OurSettings.Default_EEPROM_Settings (0xffff);
                        r = HOSTreplyMessage (Message, 0, st);
                    }
                break;
                
            case 10:    // cyclic poll
                if (len == 0)
                {
                    push_one_sample (&Message[i]);
                    r = HOSTreplyMessage (Message, N_10ms_VALUES, st);
                }
                break;
            case 11:    // cyclic poll - reading from elastic buffer - variable length response
                if (len == 4)
                {
                    Xcorr_Velocity_as_int16 = Message[i];
                    Xcorr_Corr_Coeff_as_int16 = Message[i+1];
                    Spare3 = Message[i+2];
                    FIFO_delay_D_to_Bit = Message[i+3];
                    
                    int copyBufferUsed = BufferUsed;
                    while (copyBufferUsed && (i < (HOSTmessageMaxLength - 1 - N_10ms_VALUES)))
                    {
                        pop_10ms_values_from_elastic_buffer (&Message[i]);
                        i += N_10ms_VALUES;
                        copyBufferUsed--;
                    }
                    r = HOSTreplyMessage (Message, i-3, st);
                }
                break;
            case 12:
                if (len == 2)
                {
                    int n = ProcessPoll_ISM303_INT (Message[i], Message[i+1], (int16_t*)&Message[i]);
                    r = HOSTreplyMessage (Message, n, st);
                }    
                break;
            default:
                Message[2] = 0xffff;
                r = HOSTreplyMessage (Message, 0, st); // unknown command
                break;
        }
    }
    return r;
}

uint8_t AsciiToHexNibble(char c)
{
  if (c<'0')
      return 0;
  if (c>'F')
      c -= 32;  // toupper

  if (c>'F')
      return 0;
  if ((c>'9') && (c<'A'))
      return 0;

  c -= '0';
  if (c > 9)
      c -= 7;

  return (c & 0xf);
}

void AnalyseReceivedPacket ()
{
    if (HOSTindex > 1)
    {
        HOSTindex -= 1;
        if (((HOSTmessage[1] & 0x7ff)+3) == HOSTindex)
        {
            uint16_t CRC16calc = Calc_Crc16 ((uint8_t *)HOSTmessage, HOSTindex * 2);
            uint16_t *CRC16recp = (uint16_t*)&HOSTmessage[HOSTindex];
            
            if (CRC16calc == *CRC16recp)
                haveReplyButWaitTillRS485Free = ProcessHOSTmessage ((uint16_t*)HOSTmessage, HOSTindex, HOSTresp);
        }
        else
            Serial.printf ("Rx error length <%d,%d>", HOSTindex, (HOSTmessage[0] & 0xff)+2);
     }
}

void ParseHOSTcharacter (char ch)
{
    if (ReadableProtocol)
    {
        if (ch == 'w')
        {
            HOSTnibbleCounter = 0;
            HOSTindex=0; 
        }
        else
        {
    
            if (ch == 'W')
                AnalyseReceivedPacket ();
            else
                if (((ch >= '0') && (ch <= '9')) || 
                    ((ch >= 'A') && (ch <= 'F')) || 
                    ((ch >= 'a') && (ch <= 'f')))
                {
                    HOSTnibble = AsciiToHexNibble(ch);
                    switch (HOSTnibbleCounter)
                    {
                        case 0: HOSTword = HOSTnibble;
                                HOSTnibbleCounter++;
                                break;
                        case 1:
                        case 2: HOSTword <<= 4;
                                HOSTword += HOSTnibble;
                                HOSTnibbleCounter++;
                                break;
                        case 3: HOSTword <<= 4;
                                HOSTword += HOSTnibble;
                                HOSTmessage[HOSTindex++] = HOSTword;
                                if (HOSTindex >= HOSTmessageMaxLength)
                                    HOSTindex = 0;
                                HOSTnibbleCounter= 0;
                    }
                }
                else
                {
                    HOSTindex = 0;
                    HOSTnibbleCounter= 0;
                    if (ch == 10)
                        if (haveReplyButWaitTillRS485Free)
                        {
                            Serial1.print(HOSTresp);
                            haveReplyButWaitTillRS485Free = 0;
                        }
                }
        }
    }
    else
    {
        boolean process_ch = (ch != DLE);
//Serial.printf (">%02x>", ch);
              
        if ((prev_ch == DLE) && (!DLEDLE_history))
        {
            process_ch = false;
        
            switch (ch)
            {
              case STX:
                  HOSTindex = 0;
                  HOSTnibbleCounter = 0;
                  break;
              case ETX: 

                  AnalyseReceivedPacket ();
//              Serial.printf ("\nreceived ETX\n");
                  if (haveReplyButWaitTillRS485Free)
                  {
                      Serial1.write(HOSTresp, HOSTresp_length);
                      haveReplyButWaitTillRS485Free = 0;
//              for (int i=0; i<HOSTresp_length; i++)
//                 if (i<20)
//                    Serial.printf ("%02x", HOSTresp[i]);
//              Serial.printf ("\n");
                      
                  }
                  break;
              case DLE:
                  process_ch = true;
                  break;
            }
        }
      
        if (process_ch)
        {
            switch (HOSTnibbleCounter)
            {
                case 0 : HOSTword = ch;
                    HOSTnibbleCounter += 2;
                    break;
                case 2 : HOSTword = (HOSTword << 8) | ch;
                    HOSTmessage[HOSTindex++] = HOSTword;
                    if (HOSTindex >= HOSTmessageMaxLength)
                        HOSTindex = 0;
                    HOSTnibbleCounter= 0;
                    break;
                default:
                    HOSTindex = 0;
                    HOSTnibbleCounter= 0;
                    break; 
            }
            DLEDLE_history = (ch == DLE) && (prev_ch == DLE);
//            s = s + String.format("%02x", c);
        }
        else
            DLEDLE_history = false;
        
        prev_ch = ch;
    }
}
