/*
  St_ism303dac.cpp - Library for ST sensor chip ISM303DAC, magnetormetr, accelerometer.
  Created by Sicco Dwars, Feb27, 2021.
*/

#include "Arduino.h"
#include "St_ism303dac.h"
#include <SPI.h>

St_ism303dac::St_ism303dac(String new_device_name, int CS_pin, int OE_pin, int OEpinLevel)
{
    device_name = new_device_name;
     _CS_pin = CS_pin;
     _OE_pin = OE_pin;
     _OEpinLevel = OEpinLevel;
    pinMode (CS_pin, OUTPUT);
    digitalWrite(CS_pin, HIGH);
}

void St_ism303dac::ISM303DAC_com_begin()
{

/*    if (ready_for_full_speed)
        SPI.beginTransaction(SPISettings(5000000, MSBFIRST, SPI_MODE3));  // 5 MHz CLK
    else
        SPI.beginTransaction(SPISettings(1000000, MSBFIRST, SPI_MODE3));  // 1 MHz CLK initially, turning off I2C etc

*/
        
//    SPI.beginTransaction(SPISettings(100000, MSBFIRST, SPI_MODE3));  // 100 kHz CLK initially, turning off I2C etc
    SPI.beginTransaction(SPISettings(1000000, MSBFIRST, SPI_MODE3));  // 100 kHz CLK initially, turning off I2C etc
    digitalWrite(_OE_pin, _OEpinLevel);
    
    digitalWrite(_CS_pin, LOW);
}

void St_ism303dac::ISM303DAC_com_end()
{
    digitalWrite(_CS_pin, HIGH);
    digitalWrite(_OE_pin, LOW);
    SPI.endTransaction();
}

void DisableSDO(uint32_t *Saved_SDO_PortMode_for_SDO_pin)   // SDO and SDI (MOSI and MISO) are shared pins for ISM303, don't want host SDO pushing when ISM303 must send its data to the same wire
{
    *Saved_SDO_PortMode_for_SDO_pin = CORE_PIN11_CONFIG;
    CORE_PIN11_CONFIG = 5;
}

void EnableSDO(uint32_t *Saved_SDO_PortMode_for_SDO_pin)
{
    CORE_PIN11_CONFIG = *Saved_SDO_PortMode_for_SDO_pin;
}

byte St_ism303dac::read_reg_byte (byte reg_addr)
{
    uint32_t Saved_SDO_PortMode_for_SDO_pin;

    ISM303DAC_com_begin();

    SPI.transfer(reg_addr | 0x80);

    DisableSDO(&Saved_SDO_PortMode_for_SDO_pin);
    byte data = SPI.transfer(0xff);
    EnableSDO(&Saved_SDO_PortMode_for_SDO_pin);

    ISM303DAC_com_end();
    return data;
}

uint16_t St_ism303dac::read_reg_word (byte reg_addr)
{
    uint32_t Saved_SDO_PortMode_for_SDO_pin;

    ISM303DAC_com_begin();
    SPI.transfer(reg_addr | 0x80);

    DisableSDO(&Saved_SDO_PortMode_for_SDO_pin);
       
    uint16_t data = SPI.transfer(0xff);
    data |= ((uint16_t)SPI.transfer(0xff)) << 8;

    EnableSDO(&Saved_SDO_PortMode_for_SDO_pin);

    
    ISM303DAC_com_end();
    return data;
}

int16_t St_ism303dac::read_reg_int (byte reg_addr)
{
    int16_t x;
    uint16_t *p;
    p = (uint16_t*)&x;
    *p = read_reg_word (reg_addr);
    return x;
}

void St_ism303dac::read_reg_three_words (byte reg_addr, uint16_t *words)
{
    uint32_t Saved_SDO_PortMode_for_SDO_pin;

    uint16_t data;
    uint8_t i;
    
    ISM303DAC_com_begin();
    SPI.transfer(reg_addr | 0x80);
    
    DisableSDO(&Saved_SDO_PortMode_for_SDO_pin);

    for (i=0; i<3; i++)
    {    
        data = SPI.transfer(0xff);
        data |= ((uint16_t)SPI.transfer(0xff)) << 8;
        *words++ = data;
    }

    EnableSDO(&Saved_SDO_PortMode_for_SDO_pin);
    
    ISM303DAC_com_end();
  
}

uint8_t St_ism303dac::read_reg_one_byte_then_three_words (byte reg_addr, uint16_t *words)
{
    uint32_t Saved_SDO_PortMode_for_SDO_pin;

    uint16_t data;
    uint8_t i;
    uint8_t s;
    ISM303DAC_com_begin();
    SPI.transfer(reg_addr | 0x80);
    
    DisableSDO(&Saved_SDO_PortMode_for_SDO_pin);
    s = SPI.transfer(0xff);
    for (i=0; i<3; i++)
    {    
        data = SPI.transfer(0xff);
        data |= ((uint16_t)SPI.transfer(0xff)) << 8;
        *words++ = data;
    }

    EnableSDO(&Saved_SDO_PortMode_for_SDO_pin);
    
    ISM303DAC_com_end();
    return s;
}

void St_ism303dac::write_reg_byte (byte reg_addr, byte data)
{
    ISM303DAC_com_begin();
    SPI.transfer(reg_addr & 0x7f);
    SPI.transfer(data);
    ISM303DAC_com_end();
}

void St_ism303dac::write_reg_word (byte reg_addr, uint16_t data)
{
    ISM303DAC_com_begin();
    SPI.transfer(reg_addr & 0x7f);
    SPI.transfer(data);
    SPI.transfer(data >> 8);
    ISM303DAC_com_end();
}

void St_ism303dac::init_chip(int accel_FIFO_on, int BOOT, int magnetic_AC, int get_magnetic_vector_fast, int correct_offset_rescale)
{
    _magnetic_AC = magnetic_AC;
    _get_magnetic_vector_fast = get_magnetic_vector_fast;
    _correct_offset_rescale = correct_offset_rescale;
    delay (20);
    read_reg_byte(0x0);  // dummy read to ensure SPI status is reset
 

    write_reg_byte(CTRL1_A, 0b01010000);
    write_reg_byte(FUNC_CTRL_A, 0x00);

    if (BOOT)
    {
        write_reg_byte(CTRL2_A, 0b10000001); // hard reset Accels
        delay (10);
        write_reg_byte(CFG_REG_A_M, 0b00100000); // soft reset Magnetic
        delay (2);
        write_reg_byte(CFG_REG_A_M, 0b01000000); // hard reset Magnetic
        delay (20);
    }
    
    write_reg_byte(CTRL2_A, 0b01000001); // soft reset Accels
    delay (10);
    write_reg_byte(CFG_REG_A_M, 0b00100000); // soft reset Magnetic
    delay (10);
    
    write_reg_byte(CTRL1_A, 0b01010001);
    write_reg_byte(CTRL1_A, 0b01010001);
    write_reg_byte(CTRL1_A, 0b01010001);
    write_reg_byte(CTRL2_A, 0b00000101);
    write_reg_byte(CTRL3_A, 0b00000000);
    write_reg_byte(CTRL4_A, 0b00000000);
    write_reg_byte(CTRL5_A, 0b00000000);

    if (accel_FIFO_on)
    {
        write_reg_byte(FIFO_CTRL_A, 0b11000000);
        write_reg_byte(FIFO_THS_A, 0xff);
        _FIFO_on = 1;
      
    }
    else
    {
        write_reg_byte(FIFO_CTRL_A, 0b00000000);
        _FIFO_on = 0;
    }
  
//    write_reg_byte(CFG_REG_A_M, 0b10001100);
    write_reg_byte(CFG_REG_A_M, 0b10001101);
//    write_reg_byte(CFG_REG_B_M, 0b00011011);
    write_reg_byte(CFG_REG_B_M, 0b00011000);
    write_reg_byte(CFG_REG_C_M, 0b00110000);

    write_reg_byte(INT_CRTL_REG_M, 0b10000000);

    write_reg_word (OFFSET_X_REG_L_M, 0);
    write_reg_word (OFFSET_Y_REG_L_M, 0);
    write_reg_word (OFFSET_Z_REG_L_M, 0);

    ID_a = read_reg_byte(WHO_AM_I_A);
    ID_m = read_reg_byte(WHO_AM_I_M);

    if (ID_a == 0x43)
        has_accels = 1;
    
    if (ID_m == 0x40)
        has_magneto = 1;

    ready_for_full_speed = 1;
}

void St_ism303dac::init_chip_show_results (int use_accel_FIFO, int BOOT, int magnetic_AC, int get_magnetic_vector_fast, int correct_offset_rescale)   
{
    uint8_t i;
 
    St_ism303dac::init_chip(use_accel_FIFO, BOOT, magnetic_AC, get_magnetic_vector_fast, correct_offset_rescale);

    Serial.print ("SPI chip "+device_name+" with /SS on pin "+_CS_pin+": ");    
    if (has_accels && has_magneto)
        Serial.print ("found ISM303DAC chip "+device_name);
    else
        if (!has_accels && has_magneto)
            Serial.print ("found LIS2MDLTR chip "+device_name);
        else
            Serial.print ("no ISM303 and no LIS2MDLTR found");    

    if (has_accels || has_magneto)
    {
        Serial.print (" registers are:");
        for (i=0; i<0x6f; i++)
        {
            if ((i & 0xf) == 0)
            {
                Serial.println();
                Serial.printf ("%02x: ", i);
            }   
            Serial.printf ("%02x ", read_reg_byte (i));
        }
    }
    Serial.println ();
}

void St_ism303dac::calculate_sensors_offset_and_scale (int raw_when_at_minus_100pc, int raw_when_at_plus_100pc, int *Offset, int *Scale, int FULL_SCALE_READING)
{
    int16_t v;
   
    *Offset =  (raw_when_at_plus_100pc + raw_when_at_minus_100pc) >> 1;
    *Scale = 0x4000;
    
    v = raw_when_at_plus_100pc - *Offset;
    if (v)
        *Scale = (16384L * FULL_SCALE_READING) / v;
}



int correct_offset_scale (int x, int x_offset, int x_scale)
{
   x -= x_offset;
   x *= x_scale;
   x >>= 14;      // 0x4000 = 1.0
   return x;
}

int St_ism303dac::getvalues_acceleration ()
{
    if (!has_accels)
        return 0;
    
    int n=0;
    int16_t array_of_three_int16s[3];
    
    if (_FIFO_on)
    {
        n = read_reg_byte (FIFO_SAMPLES_A);
 
        if (read_reg_byte (FIFO_SRC_A) & 0b00100000)
            n += 256;
 
        if (n)
        {
          int32_t ax_sum = 0;
          int32_t ay_sum = 0;
          int32_t az_sum = 0;
          
          int16_t k = n;
          
          while (k--)
          {
              read_reg_three_words (OUT_X_L_A, (uint16_t*)array_of_three_int16s);
              
              ax_sum += array_of_three_int16s[0];
              ay_sum += array_of_three_int16s[1];
              az_sum += array_of_three_int16s[2];
          }
          acc.x = ax_sum;
          acc.y = ay_sum;
          acc.z = az_sum;
          acc.nav = n;
        }
    }
    
    if (!n)
    {
        status_a = read_reg_one_byte_then_three_words (STATUS_A, (uint16_t*)array_of_three_int16s);
        if (status_a & 1)
        {
            acc.x = array_of_three_int16s[0];
            acc.y = array_of_three_int16s[1];
            acc.z = array_of_three_int16s[2];
            acc.nav = 1;
        }
    }
    return (n);
}

int St_ism303dac::getvalues_magnetic ()
{
    int16_t array_of_three_int16s[3];
      
    if (!has_magneto)
        return 0;        
    
    status_m = read_reg_one_byte_then_three_words (STATUS_REG_M, (uint16_t*)array_of_three_int16s);
    if (status_m & 1)
        mag.x = array_of_three_int16s[0];
    if (status_m & 2)
        mag.y = array_of_three_int16s[1];
    if (status_m & 4)
        mag.z = array_of_three_int16s[2];

    mag.nav = 1;
    
    write_reg_byte(CFG_REG_A_M, 0b10001101);
    
    return (status_m & 7);
}

int St_ism303dac::getvalue_temperature ()
{
    if (has_accels)
        temperature = 25 + read_reg_byte(OUT_T_A);
    else
        if (has_magneto)
        {
            int16_t t = read_reg_int (TEMP_OUT_L_REG);
            temperature = 25 + (t >> 3);
        }
        else
            temperature = 0;
        
    return temperature;
}

void St_ism303dac::copy_and_clear_integrator ()
{
    int r = acc_int.nav;
    if (r)
    {
        acc_f.x = acc_int.x / r;
        acc_f.y = acc_int.y / r;
        acc_f.z = acc_int.z / r;
        acc_f.nav = r;
    }
    
    r = mag_int.nav;
    if (r)
    {
        mag_f.x = mag_int.x / r;
        mag_f.y = mag_int.y / r;
        mag_f.z = mag_int.z / r;
        mag_f.nav = r;
    }
}

void St_ism303dac::add_to_integrator ()
{
    if (flag_get_averages)
    {
        copy_and_clear_integrator ();
        acc_int.x = acc.x;
        acc_int.y = acc.y;
        acc_int.z = acc.z;
        acc_int.nav = acc.nav;
        mag_int.x = mag.x;
        mag_int.y = mag.y;
        mag_int.z = mag.z;
        mag_int.nav = 1;
        flag_get_averages = 0;
    }
    else
    {
        acc_int.x += acc.x;
        acc_int.y += acc.y;
        acc_int.z += acc.z;
        acc_int.nav += acc.nav;
        mag_int.x += mag.x;
        mag_int.y += mag.y;
        mag_int.z += mag.z;
        mag_int.nav += 1;
    }
}

void low_pass_filter (int new_value, int nav, int *filter_out, float filter_constant)
{
    float s = (float)*filter_out;
    
    if (nav)
        s = s * (1.0 - filter_constant) + (float)((new_value << 4)/nav) * filter_constant;

    *filter_out = (int)s;
}

void low_pass_filter_axis (axis_t *axis, axis_t *filtered_axis, float filter_constant)
{
    low_pass_filter (axis->x, axis->nav, &filtered_axis->x, filter_constant);
    low_pass_filter (axis->y, axis->nav, &filtered_axis->y, filter_constant);
    low_pass_filter (axis->z, axis->nav, &filtered_axis->z, filter_constant);
}

void check_min_max (int new_value, int *min_value, int *max_value)
{
    if (new_value < *min_value)
        *min_value = new_value;
    if (new_value > *max_value)
        *max_value = new_value;
}

void check_min_max_axis (axis_t *axis, axis_t *axis_extremes_min, axis_t *axis_extremes_max) 
{
      check_min_max (axis->x, &axis_extremes_min->x, &axis_extremes_max->x);
      check_min_max (axis->y, &axis_extremes_min->y, &axis_extremes_max->y);
      check_min_max (axis->z, &axis_extremes_min->z, &axis_extremes_max->z);
}

#define filter 0.05

void St_ism303dac::getAllSensorValues (bool toggle_bit)
{
    St_ism303dac::getvalues_acceleration();
    St_ism303dac::getvalues_magnetic ();
    St_ism303dac::getvalue_temperature ();

    low_pass_filter_axis (&acc, &filtered_acc, filter);
    low_pass_filter_axis (&mag, &filtered_mag, filter);
    
    if (flag_reset_extremes)
    {
        flag_reset_extremes = 0;
        extremes.maximum.acc.x = filtered_acc.x;
        extremes.maximum.acc.y = filtered_acc.y;
        extremes.maximum.acc.z = filtered_acc.z;
        
        extremes.maximum.mag.x = filtered_mag.x;
        extremes.maximum.mag.y = filtered_mag.y;
        extremes.maximum.mag.z = filtered_mag.z;

        extremes.minimum.acc.x = filtered_acc.x;
        extremes.minimum.acc.y = filtered_acc.y;
        extremes.minimum.acc.z = filtered_acc.z;
        
        extremes.minimum.mag.x = filtered_mag.x;
        extremes.minimum.mag.y = filtered_mag.y;
        extremes.minimum.mag.z = filtered_mag.z;
    }
    else
    {
        check_min_max_axis (&filtered_acc, &extremes.minimum.acc, &extremes.maximum.acc);
        check_min_max_axis (&filtered_mag, &extremes.minimum.mag, &extremes.maximum.mag);
    }

    if (_correct_offset_rescale)
    {
        acc.x = correct_offset_scale (acc.x, acc_calibrate_offsets.x, acc_calibrate_scales.x);
        acc.y = correct_offset_scale (acc.y, acc_calibrate_offsets.y, acc_calibrate_scales.y);
        acc.z = correct_offset_scale (acc.z, acc_calibrate_offsets.z, acc_calibrate_scales.z);
        
        mag.x = correct_offset_scale (mag.x, mag_calibrate_offsets.x, mag_calibrate_scales.x);
        mag.y = correct_offset_scale (mag.y, mag_calibrate_offsets.y, mag_calibrate_scales.y);
        mag.z = correct_offset_scale (mag.z, mag_calibrate_offsets.z, mag_calibrate_scales.z);
    }

    if (_get_magnetic_vector_fast)
    {
        float MX =  mag.x;
        float MY =  mag.y;
        float MZ =  mag.z;
        if (_magnetic_AC)
        {
            MX -= mag_previous.x;
            MY -= mag_previous.y;
            MZ -= mag_previous.z;
        }   
        mag_previous.x = mag.x;
        mag_previous.y = mag.y;
        mag_previous.z = mag.z;
            
        mag_magnitude_AC_fast =  sqrtf( (MX*MX) + (MY*MY) + (MZ*MZ));
    }
    
    if (toggle_bit && _magnetic_AC)
    {
        mag.x = -mag.x; 
        mag.y = -mag.y; 
        mag.z = -mag.z; 
    }
    add_to_integrator ();
}

void St_ism303dac::printAllSensorValues (int mode_bit_flags)
{
    int len = device_name.length();
    int k = 0;
    char *lst = txt_st;

    if (!(has_magneto || has_accels))
    {
        *lst = 0;
        return;
    }
    
    *lst++ = ' ';
    while (len--)
        *lst++ = device_name.charAt(k++);

    if (mode_bit_flags & ACC_XYZ)
        if (has_accels)
        {
            sprintf (lst, " aXYZ(%d): %5d %5d %5d", acc_f.nav, acc_f.x, acc_f.y, acc_f.z);
            lst += strlen(lst);
        }
        
    if (mode_bit_flags & MAG_XYZ)
        if (has_magneto)
        {
            sprintf (lst, " mXYZ(%d): %5d %5d %5d", mag_f.nav, mag_f.x, mag_f.y, mag_f.z);
            lst += strlen(lst);
        }
    
    if (mode_bit_flags & MAG_MAGNITUDE)
    {
        float MX2 =  mag_f.x *  mag_f.x;
        float MY2 =  mag_f.y *  mag_f.y;
        float MZ2 =  mag_f.z *  mag_f.z;
        
        float mag_magnitude = sqrtf(MX2 + MY2 + MZ2);
        sprintf (lst, " %1.1f", mag_magnitude);
        lst += strlen(lst);
    }

    if (mode_bit_flags & TEMPERATURE)
    {
        sprintf (lst, " t=%2d", temperature);
        lst += strlen(lst);
    }
}
