/*
  St_ism303dac.h - Library for ST sensor chip ISM303DAC, magnetormetr, accelerometer.
  Created by Sicco Dwars, Feb27, 2021.

*/
#ifndef St_ism303dac_h
#define St_ism303dac_h

#include "Arduino.h"

#define ACC_XYZ       0x01
#define MAG_XYZ       0x02
#define TEMPERATURE   0x04
#define MAG_MAGNITUDE 0x08

#define ACC_100pc 16384
#define MAG_100pc 300

// ST IMS303DAC Chip Registers
#define RESERVED_00     0x00
#define RESERVED_01     0x01
#define RESERVED_02     0x02
#define RESERVED_03     0x03
#define RESERVED_04     0x04
#define RESERVED_05     0x05
#define RESERVED_06     0x06
#define RESERVED_07     0x07
#define RESERVED_08     0x08
#define RESERVED_09     0x09
#define RESERVED_0A     0x0a
#define RESERVED_0B     0x0b

#define Module_8bit_A   0x0c
#define RESERVED_0D     0x0d
#define RESERVED_0E     0x0e
#define WHO_AM_I_A      0x0f
#define RESERVED_10     0x10
#define RESERVED_11     0x11
#define RESERVED_12     0x12
#define RESERVED_13     0x13
#define RESERVED_14     0x14
#define RESERVED_15     0x15
#define RESERVED_16     0x16
#define RESERVED_17     0x17
#define RESERVED_18     0x18
#define RESERVED_19     0x19
#define RESERVED_1A     0x1a
#define RESERVED_1B     0x1b
#define RESERVED_1C     0x1c
#define RESERVED_1D     0x1d
#define RESERVED_1E     0x1e
#define RESERVED_1F     0x1f

#define RESERVED_11     0x11
#define RESERVED_11     0x11
#define RESERVED_11     0x11
#define RESERVED_11     0x11

#define CTRL1_A         0x20
#define CTRL2_A         0x21
#define CTRL3_A         0x22
#define CTRL4_A         0x23
#define CTRL5_A         0x24
#define FIFO_CTRL_A     0x25
#define OUT_T_A         0x26
#define STATUS_A        0x27
#define OUT_X_L_A       0x28
#define OUT_X_H_A       0x29
#define OUT_Y_L_A       0x2A
#define OUT_Y_H_A       0x2B
#define OUT_Z_L_A       0x2C
#define OUT_Z_H_A       0x2D
#define FIFO_THS_A      0x2E
#define FIFO_SRC_A      0x2F
#define FIFO_SAMPLES_A  0x30

#define TAP_6D_THS_A    0x31
#define INT_DUR_A       0x32
#define WAKE_UP_THS_A   0x33
#define WAKE_UP_DUR_A   0x34
#define FREE_FALL_A     0x35
#define STATUS_DUP_A    0x36
#define WAKE_UP_SRC_A   0x37
#define TAP_SRC_A       0x38
#define _6D_SRC_A       0x39
#define RESERVED_3A     0x3a
#define RESERVED_3B     0x3b
#define RESERVED_3C     0x3c
#define RESERVED_3D     0x3d
#define RESERVED_3E     0x3e
#define FUNC_CTRL_A     0x3f

#define RESERVED_40     0x40
#define RESERVED_41     0x41
#define RESERVED_42     0x42
#define RESERVED_43     0x43
#define RESERVED_44     0x44
#define OFFSET_X_REG_L_M    0x45
#define OFFSET_X_REG_H_M    0x46
#define OFFSET_Y_REG_L_M    0x47
#define OFFSET_Y_REG_H_M    0x48
#define OFFSET_Z_REG_L_M    0x49
#define OFFSET_Z_REG_H_M    0x4A
#define RESERVED_4B     0x4b
#define RESERVED_4C     0x4c
#define RESERVED_4D     0x4d
#define RESERVED_4E     0x4e
#define WHO_AM_I_M      0x4f

#define RESERVED_50     0x50
#define RESERVED_51     0x51
#define RESERVED_52     0x52
#define RESERVED_53     0x53
#define RESERVED_54     0x54
#define RESERVED_55     0x55
#define RESERVED_56     0x56
#define RESERVED_57     0x57
#define RESERVED_58     0x58
#define RESERVED_59     0x59
#define RESERVED_5A     0x5a
#define RESERVED_5B     0x5b
#define RESERVED_5C     0x5c
#define RESERVED_5D     0x5d
#define RESERVED_5E     0x5e
#define RESERVED_5F     0x5f
#define CFG_REG_A_M     0x60
#define CFG_REG_B_M     0x61
#define CFG_REG_C_M     0x62
#define INT_CRTL_REG_M  0x63
#define INT_SOURCE_REG_M    0x64
#define INT_THS_L_REG_M 0x65
#define INT_THS_H_REG_M 0x66
#define STATUS_REG_M    0x67
#define OUTX_L_REG_M    0x68
#define OUTX_H_REG_M    0x69
#define OUTY_L_REG_M    0x6a
#define OUTY_H_REG_M    0x6b
#define OUTZ_L_REG_M    0x6c
#define OUTZ_H_REG_M    0x6d
//#define RESERVED_6E     0x6e
//#define RESERVED_6F     0x6f
#define TEMP_OUT_L_REG  0x6e
#define TEMP_OUT_H_REG  0x6f

typedef struct 
{
  int x;
  int y;
  int z;
  int nav;
} axis_t;

typedef struct
{
  axis_t acc;
  axis_t mag;
} acc_mag_axis_t;

typedef struct
{
  acc_mag_axis_t maximum;
  acc_mag_axis_t minimum;
} acc_mag_axis_extremes_t;

class St_ism303dac
{
  public:
    
    St_ism303dac(String new_device_name, int CS_pin, int OE_pin, int OEpinLevel);
    
    byte read_reg_byte (byte reg_addr);
    uint16_t read_reg_word (byte reg_addr);
    int16_t read_reg_int (byte reg_addr);
    void read_reg_three_words (byte reg_addr, uint16_t *words);
    uint8_t read_reg_one_byte_then_three_words (byte reg_addr, uint16_t *words);
    void write_reg_byte (byte reg_addr, byte data);
    void write_reg_word (byte reg_addr, uint16_t data);

    void init_chip(int accel_FIFO_on, int BOOT, int magnetic_AC, int get_magnetic_vector_fast, int correct_offset_rescale);
    void init_chip_show_results (int use_accel_FIFO, int BOOT, int magnetic_AC, int get_magnetic_vector_fast, int correct_offset_rescale);   
 
    void calculate_sensors_offset_and_scale (int raw_when_at_minus_100pc, int raw_when_at_plus_100pc, int *Offset, int *Scale, int FULL_SCALE_READING);

    int getvalues_acceleration();
    int getvalues_magnetic ();
    int getvalue_temperature ();
    void getAllSensorValues (bool toggle_bit);

    void printAllSensorValues (int mode_bit_flags);

    String device_name = "";
    
	  uint8_t status_a = 0;
	  uint8_t status_m = 0;
    
    axis_t mag_f;
    axis_t acc_f;
    
    bool flag_get_averages = 0;

    acc_mag_axis_extremes_t extremes;
    bool flag_reset_extremes = 1;

	  int8_t temperature;

    bool has_accels = 0;
    bool has_magneto = 0;

    bool _correct_offset_rescale = 0;
  
    void ISM303DAC_com_begin();
    void ISM303DAC_com_end();
    char txt_st[256];

    float mag_magnitude_AC_fast = 0.0;

    axis_t mag_calibrate_offsets = {0,0,0,1};
    axis_t mag_calibrate_scales = {0x4000, 0x4000, 0x4000, 1};
    
    axis_t acc_calibrate_offsets = {0,0,0,1};
    axis_t acc_calibrate_scales = {0x4000, 0x4000, 0x4000, 1};
 
    axis_t filtered_mag;
    axis_t filtered_acc;

    
  private:
    void copy_and_clear_integrator ();
    void add_to_integrator ();
  
    int _CS_pin;
    int _OE_pin;
    int _OEpinLevel;
    int _FIFO_on;
    int _magnetic_AC;
    uint8_t ID_a, ID_m;
    bool ready_for_full_speed = 0;

    axis_t mag_int;
    axis_t acc_int;

    axis_t mag;
    axis_t acc;



    bool _get_magnetic_vector_fast = 0;
    axis_t mag_previous;


};

#endif
