#include <BufferedPrint.h>
#include <FreeStack.h>
#include <MinimumSerial.h>
#include <RingBuf.h>
#include <SdFat.h>
#include <SdFatConfig.h>
#include <sdios.h>

#include <TimeLib.h>

#define forward_is_anti_clock_wise  // comment out this line for clockwise = forward = positive speed, position

#ifdef forward_is_anti_clock_wise
    #define default_forward_def_motor HIGH
    #define encoder_reverse    
#else
    #define default_forward_def_motor LOW
    #undef encoder_reverse
#endif

#include <SdFatConfig.h>
#include <SdFat.h>

SdFs sd;
FsFile txtFile;

#include "EEPROM_settings.h"

#include "i2c_driver.h"
#include "imx_rt1060_i2c_driver.h"

#include "INA260.h" 

#include "MCP4728.h"

#include "St_ism303dac.h"
#include "dircalcs.h"

#include "QuadEncoder.h"

#include <ADC.h>
#include <ADC_util.h>

#include <imxrt.h>

#include <util/atomic.h>

#include "HostProtocol.h"
#include "Enc50t_Protocol.h"

EEPROM_settings OurSettings;

IntervalTimer myTimer_10ms;
IntervalTimer myTimer_100ms;

IMX_RT1060_I2CMaster& I2C_master = Master;

INA260 ina260_motor = INA260(); 
INA260 ina260_coil = INA260();

MCP4728 DAC_MCP4728;

volatile char logFileTypeString[20] = "CANOP02v04";
volatile char versionString[40] = "v001.010 by SD";

volatile uint16_t global_DAC_CH_A_value=0, global_DAC_CH_B_value=2001, global_DAC_CH_C_value=2002, global_DAC_CH_D_value=0;
volatile bool flag_new_dac_values = 0;

volatile int SetpointStep = 10;
volatile bool MotorEnable = 1;
volatile bool MotorTestScript = 0;

volatile int modulate_AC_on = 1;
//volatile int modulate_AC_on = 0;

const int OE_ST2378_Pin = 5;

/*const int CS_pin_for_MA = 14;
const int CS_pin_for_MB = 15;
const int CS_pin_for_MC = 9;
const int CS_pin_for_MD = 17;
*/
const int CS_pin_for_MD = 14;
const int CS_pin_for_MC = 15;
const int CS_pin_for_MB = 9;
const int CS_pin_for_MA = 17;

St_ism303dac ISM303_A ("MA", CS_pin_for_MA, OE_ST2378_Pin, 0);    
St_ism303dac ISM303_B ("MB", CS_pin_for_MB, OE_ST2378_Pin, 0);   
St_ism303dac ISM303_C ("MC", CS_pin_for_MC, OE_ST2378_Pin, 0);     
St_ism303dac ISM303_D ("MD", CS_pin_for_MD, OE_ST2378_Pin, 0);   

St_ism303dac ISM303_INT ("ISM303_INT", 10, OE_ST2378_Pin, 1);  // cs3

float MA, MB, MC, MD, ME;
float motor_current;

int show_magnetic_extremes_ISM303_INT = 0;
int show_acceleration_extremes_ISM303_INT = 0;

#ifdef encoder_reverse
    QuadEncoder MaxonMotorEncoderOnTeensy4 (1, 4, 3);   // swap A and B inputs for reverse direction as a positive number
#else
    QuadEncoder MaxonMotorEncoderOnTeensy4 (1, 3, 4);
#endif

volatile bool FilteredBinaryInputPremexReedSwitch = 0;
volatile int IntInputPremexReedSwitch = 10000;
volatile bool syncOnPremex = 1;

volatile bool zero_rt = 0;
volatile int32_t speed_encoder = 0, pos_encoder = 0, previous_pos_encoder = 0;
volatile int32_t encoder_pos = 0, encoder_speed = 0;

volatile int32_t pos_encoder_av = 0;
volatile int32_t speed_encoder_av = 0;
volatile int32_t encoder_nav = 0;
bool flag_clear_encoder_averages = 0;

const uint32_t EncoderPulsesPerMotorRev = 500; 

const uint32_t GearBoxRatioN = 388125;    // Maxon Gearhead - GPX16HP 83:1   -  so it actually is 1:82.70296185808651
const uint32_t GearBoxRatioD = 4693;
//const float GearBoxRatio = 82.70296185808651;

const int32_t ShaftSyncTurn = EncoderPulsesPerMotorRev * GearBoxRatioN * GearBoxRatioD;  // so many encoder pulses until we're guaranteed back in sync
const int32_t EncoderPulsesPerDiverterRev = EncoderPulsesPerMotorRev * GearBoxRatioN / GearBoxRatioD;

volatile int pos_encoder_int32_sync = 0;

volatile long int pos_in_one_turn;
volatile float exit_shaft_orientation;
volatile float RPM_exit_shaft;
volatile float exit_shaft_orientation_at_Premex_edge = 0.0;

const int ledPin = 13;

const int LDAC_Pin = 20;

const int PENABLE_Pin = 32;
const int H_BRIDGE_1_Pin = 35;
const int H_BRIDGE_2_Pin = 36;

const int RS485_DIR1_Pin = 2;
const int RS485_DIR2_Pin = 6;

const int BinaryInputPremexReedSwitch_Pin = 38; // previously MOT_READY, disconnect R41
volatile int32_t duration_PremexReedSwitch_high = -1;
volatile int32_t duration_PremexReedSwitch_low = -1;
volatile int32_t prev_duration_PremexReedSwitch_high = -1;
volatile int32_t prev_duration_PremexReedSwitch_low = -1;
volatile uint16_t ShaftEncoderAndPremexReedSwitch = 0;

volatile bool passed360to0 = 0;
volatile int QuadrantWeReIn, prev_QuadrantWeReIn = -1;

enum {INCODER_SSI, ENCODER_ABZ, ISM303_MAGNETIC_TOOLFACE, ISM303_GRAVITATIONAL_TOOLFACE, ISM303_AUTO_TOOLFACE, SIMULATE};
volatile uint8_t SourceForBitOrientation = INCODER_SSI;

enum {MAGNETIC, GRAVITATIONAL};
volatile int TF_auto_mode = GRAVITATIONAL;

volatile float inclination = 0.0;
volatile float total_grav = 0.0;
volatile float total_magn = 0.0;
volatile float toolface_g = 0.0;
volatile float toolface_m = 0.0;    
volatile float azimuth = 0.0;
volatile float dipangle = 0.0;
volatile float toolface_auto = 0.0;    
 
const int ina260_motor_alertPin = 37; 
const int ina260_coil_alertPin = 26;

const int motorEnablePin = 39;
const int motorDirReversePin = 40;

volatile int ledState = 0;
volatile int BlinkDurationIn10msCounts = 20;

bool toggle_bit = 0;

volatile int Count_10ms_ticks = 0;
volatile int Count2_10ms_ticks = 0;
volatile int Count_10ms_ticks_no_Alert_isr_motor = 0;
volatile int Count_10ms_ticks_no_Alert_isr_coil = 0;
volatile int Count_10ms_ticks_no_100ms_isr = 0;
volatile int Count_10ms_missed_loop_cycles = 0;
volatile int Count_10ms_missed_SD_write = 0;

volatile bool flag_1seconds_trigger = 0;
volatile bool flag_next_1seconds_trigger = 0;
volatile int Count_100ms_ticks_for_flag_1seconds_trigger = 0;

volatile int Count_100ms_ticks = 0;
volatile int Count_100ms_ticks_no_Alert_isr_motor = 0;
volatile int Count_100ms_ticks_no_Alert_isr_coil = 0;

volatile int motor_test_script_second = 0;
volatile int motor_test_script_step = 0;

volatile int UpTimeSeconds = 0;

volatile float current_integrator_motor = 0.0, voltage_integrator_motor = 0.0, power_integrator_motor = 0.0;
volatile int nav_current_motor = 0;
volatile bool flag_clear_current_averages_motor = 0;
volatile float current_integrator_coil = 0.0, voltage_integrator_coil = 0.0, power_integrator_coil = 0.0;
volatile int nav_current_coil = 0;
volatile bool flag_clear_current_averages_coil = 0;

ADC *adc = new ADC(); 

#define L2_BUF_SIZE     14
#define BUF_SIZE        ((uint32_t)1 << L2_BUF_SIZE)
#define BUF_SIZE_MASK   (BUF_SIZE-1)

#define WRITE_TO_SD_CHUNK_SIZE (BUF_SIZE >> 1)

volatile uint8_t Buffer_a [BUF_SIZE];
volatile uint32_t Buffer_a_write_pointer = 0;
volatile uint32_t Buffer_a_read_pointer = 0;
volatile uint32_t Buffer_a_used = 0;

volatile uint8_t Buffer_b [BUF_SIZE];
volatile uint32_t Buffer_b_write_pointer = 0;
volatile uint32_t Buffer_b_read_pointer = 0;
volatile uint32_t Buffer_b_used = 0;

volatile char old_log_filename_a[40] = "";
volatile char old_log_filename_b[40] = "";

volatile int32_t n_100ms_samples_since_boot = 0;
volatile int32_t n_10ms_samples_since_boot = 0;

volatile int debug = 0;
volatile uint32_t SystemStatusFlags = 0;

static uint8_t Serial1WriteBuffer[4096];
static uint8_t Serial1ReadBuffer[4096];

#define flag_Buffer_a_overflow                  0x0001
#define flag_Buffer_a_over_writing              0x0002
#define flag_in_prepareSDcard                   0x0004
#define flag_overasking_loop_actions            0x0008

#define flag_overasking_100ms_actions           0x0010
#define flag_write_to_log_file                  0x0020
#define flag_write_to_log_file_SDcard_actions   0x0040
#define flag_in_loop                            0x0080

#define flag_in_add_string_to_buffer_a_for_SDcard_file_writes 0x0100
#define flag_in_isr_100ms                       0x0200
#define flag_in_isr_alertPinAction              0x0400
#define flag_missed_Alert_isr_motor             0x0800

#define flag_missed_Alert_isr_coil              0x1000
#define flag_missed_100ms_isr                   0x2000
#define flag_missed_loop_cycles                 0x4000
#define flag_missed_SD_write                    0x8000

#define flag_did_loop_in_past_200ms             0x10000
#define flag_did_INA260_motor_in_past_200ms     0x20000
#define flag_did_INA260_coil_in_past_200ms      0x40000
#define flag_did_SD_write_in_past_second        0x80000

#define flag_write_to_SD_fails                  0x100000
#define flag_Buffer_b_overflow                  0x200000


volatile bool mutex = 0;

typedef struct
{
    float P_gain; 
    float I_gain;
    float max_out;
    float min_out;
    float P_contribution;
    float I_contribution;
    bool clipping_hi;
    bool clipping_lo;
    float PI_out;
} PI_Controller_t;

volatile PI_Controller_t PositionController;
volatile PI_Controller_t SpeedController;
volatile float PosCntrl_P_gain = 10.0;
volatile float PosCntrl_I_gain = 0.005;
volatile float max_PWM_demanded = 2000.0;    // DAC counts for RPM setpoint to Maxon speed controller
volatile float min_PWM_demanded = -2000.0;   // -DAC counts for RPM setpoint to Maxon speed controller

volatile float SpeedCntrl_P_gain = 2.0;
volatile float SpeedCntrl_I_gain = 0.5;
volatile float max_torque_demand = 2000.0;    // DAC counts for RPM setpoint to Maxon speed controller
volatile float min_torque_demand = -2000.0;   // -DAC counts for RPM setpoint to Maxon speed controller

volatile float ToolFace50t_TriggerOffsetSetting = 0.0;
volatile float ToolFace50t_Encoder = 0.0;
volatile float prev_Shifted50t_encoder = 0.0;
volatile float Shifted50t_encoder = 0.0;
volatile float setpoint_diverter_orientation = 0.0;
volatile float measured_diverter_orientation = 0.0;
volatile float orientation_error = 0.0;

volatile float orientation_step= 0.0;
volatile float prev_measured_diverter_orientation = 0.0;
volatile float integrated_orientation_steps = 0.0;

volatile float speed_error = 0.0;
volatile float setpoint_RPM = 20.0;
volatile float setpoint_RPMfast = 300.0;

volatile float setpoint_RPM_exit_shaft = 0.0;

volatile int PWM_demanded = 0;

volatile bool Position_Control = 0;
volatile bool Speed_Control = 0;
volatile bool prevSpeed_Control = 0;

volatile int control_and_trigger_mode = 0;

volatile int TautoTrigger = 0;
volatile int Timer_TautoTrigger = 0;

void setup_PI_controller (volatile PI_Controller_t *PIctrl, float P_gain, float I_gain, float min_out, float max_out)
{
    PIctrl->P_gain = P_gain;
    PIctrl->I_gain = I_gain;
    PIctrl->min_out = min_out;
    PIctrl->max_out = max_out;
    PIctrl->P_contribution = 0.0;
    PIctrl->I_contribution = 0.0;
    PIctrl->PI_out = 0.0;
    PIctrl->clipping_hi = 0;
    PIctrl->clipping_lo = 0;
}

float PI_control_step (float epsilon, volatile PI_Controller_t *C)
{  
    float Q;
    
    C->P_contribution = epsilon * C->P_gain;
    C->I_contribution += epsilon * C->I_gain;    
    
    Q = C->P_contribution + C->I_contribution;
    
    C->clipping_hi = (Q > C->max_out);
    C->clipping_lo = (Q < C->min_out);
    
    if (C->clipping_hi)
        Q = C->max_out;

    if (C->clipping_lo)
        Q = C->min_out;

    if (C->clipping_hi || C->clipping_lo)
        C->I_contribution = Q - C->P_contribution;
    
    C->PI_out = Q;
        
    return Q;
}


void set_offsets_scales (St_ism303dac *ISM303)
{
    ISM303_INT.calculate_sensors_offset_and_scale (OurSettings.SettingsBuffer.ISM303_INT_acc_x_minus100pcReading, OurSettings.SettingsBuffer.ISM303_INT_acc_x_plus100pcReading, &ISM303->acc_calibrate_offsets.x, &ISM303->acc_calibrate_scales.x, ACC_100pc);
    ISM303_INT.calculate_sensors_offset_and_scale (OurSettings.SettingsBuffer.ISM303_INT_acc_y_minus100pcReading, OurSettings.SettingsBuffer.ISM303_INT_acc_y_plus100pcReading, &ISM303->acc_calibrate_offsets.y, &ISM303->acc_calibrate_scales.y, ACC_100pc);
    ISM303_INT.calculate_sensors_offset_and_scale (OurSettings.SettingsBuffer.ISM303_INT_acc_z_minus100pcReading, OurSettings.SettingsBuffer.ISM303_INT_acc_z_plus100pcReading, &ISM303->acc_calibrate_offsets.z, &ISM303->acc_calibrate_scales.z, ACC_100pc);

    ISM303_INT.calculate_sensors_offset_and_scale (OurSettings.SettingsBuffer.ISM303_INT_mag_x_minus100pcReading, OurSettings.SettingsBuffer.ISM303_INT_mag_x_plus100pcReading, &ISM303->mag_calibrate_offsets.x, &ISM303->mag_calibrate_scales.x, MAG_100pc);
    ISM303_INT.calculate_sensors_offset_and_scale (OurSettings.SettingsBuffer.ISM303_INT_mag_y_minus100pcReading, OurSettings.SettingsBuffer.ISM303_INT_mag_y_plus100pcReading, &ISM303->mag_calibrate_offsets.y, &ISM303->mag_calibrate_scales.y, MAG_100pc);
    ISM303_INT.calculate_sensors_offset_and_scale (OurSettings.SettingsBuffer.ISM303_INT_mag_z_minus100pcReading, OurSettings.SettingsBuffer.ISM303_INT_mag_z_plus100pcReading, &ISM303->mag_calibrate_offsets.z, &ISM303->mag_calibrate_scales.z, MAG_100pc);
}

void AcceptISM303_INT_ResetExtremes()
{
    ISM303_INT.flag_reset_extremes = 1;
}

void AcceptISM303_INT_MagExtremes ()
{
    OurSettings.SettingsBuffer.ISM303_INT_mag_x_minus100pcReading = ISM303_INT.extremes.minimum.mag.x / 16;
    OurSettings.SettingsBuffer.ISM303_INT_mag_x_plus100pcReading = ISM303_INT.extremes.maximum.mag.x / 16;
    OurSettings.SettingsBuffer.ISM303_INT_mag_y_minus100pcReading = ISM303_INT.extremes.minimum.mag.y / 16;
    OurSettings.SettingsBuffer.ISM303_INT_mag_y_plus100pcReading = ISM303_INT.extremes.maximum.mag.y / 16;
    OurSettings.SettingsBuffer.ISM303_INT_mag_z_minus100pcReading = ISM303_INT.extremes.minimum.mag.z / 16;
    OurSettings.SettingsBuffer.ISM303_INT_mag_z_plus100pcReading = ISM303_INT.extremes.maximum.mag.z / 16;
    set_offsets_scales (&ISM303_INT);
    OurSettings.Write_EEPROM_Settings ();
}

void AcceptISM303_INT_AccExtremes ()
{
    OurSettings.SettingsBuffer.ISM303_INT_acc_x_minus100pcReading = ISM303_INT.extremes.minimum.acc.x / 16;
    OurSettings.SettingsBuffer.ISM303_INT_acc_x_plus100pcReading = ISM303_INT.extremes.maximum.acc.x / 16;
    OurSettings.SettingsBuffer.ISM303_INT_acc_y_minus100pcReading = ISM303_INT.extremes.minimum.acc.y / 16;
    OurSettings.SettingsBuffer.ISM303_INT_acc_y_plus100pcReading = ISM303_INT.extremes.maximum.acc.y / 16;
    OurSettings.SettingsBuffer.ISM303_INT_acc_z_minus100pcReading = ISM303_INT.extremes.minimum.acc.z / 16;
    OurSettings.SettingsBuffer.ISM303_INT_acc_z_plus100pcReading = ISM303_INT.extremes.maximum.acc.z / 16;
    set_offsets_scales (&ISM303_INT);
    OurSettings.Write_EEPROM_Settings ();
}
   
void AcceptISM303_Defaults ()
{
    OurSettings.Default_EEPROM_Settings (1);
    set_offsets_scales (&ISM303_INT);
}

enum {NOP, CLEAR_MIN_MAX, ACCEPT_PEAK_FOR_FLAGGED_SENSORS, ACCEPT_MAG, ACCEPT_ACC, DEFAULTS};

int ProcessPoll_ISM303_INT (int comm, int param, int16_t *resp)
{
    switch (comm)   
    {
        case NOP:
            break;
        case CLEAR_MIN_MAX:
            AcceptISM303_INT_ResetExtremes();
            break;
        case ACCEPT_PEAK_FOR_FLAGGED_SENSORS:
            if (param & 0x1)
                OurSettings.SettingsBuffer.ISM303_INT_mag_x_minus100pcReading = ISM303_INT.extremes.minimum.mag.x / 16;
            if (param & 0x2)
                OurSettings.SettingsBuffer.ISM303_INT_mag_x_plus100pcReading = ISM303_INT.extremes.maximum.mag.x / 16;
            if (param & 0x4)
                OurSettings.SettingsBuffer.ISM303_INT_mag_y_minus100pcReading = ISM303_INT.extremes.minimum.mag.y / 16;
            if (param & 0x8)
                OurSettings.SettingsBuffer.ISM303_INT_mag_y_plus100pcReading = ISM303_INT.extremes.maximum.mag.y / 16;
            if (param & 0x10)
                OurSettings.SettingsBuffer.ISM303_INT_mag_z_minus100pcReading = ISM303_INT.extremes.minimum.mag.z / 16;
            if (param & 0x20)
                OurSettings.SettingsBuffer.ISM303_INT_mag_z_plus100pcReading = ISM303_INT.extremes.maximum.mag.z / 16;
            if (param & 0x40)
                OurSettings.SettingsBuffer.ISM303_INT_acc_x_minus100pcReading = ISM303_INT.extremes.minimum.acc.x / 16;
            if (param & 0x80)
                OurSettings.SettingsBuffer.ISM303_INT_acc_x_plus100pcReading = ISM303_INT.extremes.maximum.acc.x / 16;
            if (param & 0x100)
                OurSettings.SettingsBuffer.ISM303_INT_acc_y_minus100pcReading = ISM303_INT.extremes.minimum.acc.y / 16;
            if (param & 0x200)
                OurSettings.SettingsBuffer.ISM303_INT_acc_y_plus100pcReading = ISM303_INT.extremes.maximum.acc.y / 16;
            if (param & 0x400)
                OurSettings.SettingsBuffer.ISM303_INT_acc_z_minus100pcReading = ISM303_INT.extremes.minimum.acc.z / 16;
            if (param & 0x800)
                OurSettings.SettingsBuffer.ISM303_INT_acc_z_plus100pcReading = ISM303_INT.extremes.maximum.acc.z / 16;
            set_offsets_scales (&ISM303_INT);
            OurSettings.Write_EEPROM_Settings ();

            break;
        case ACCEPT_MAG:
            AcceptISM303_INT_MagExtremes ();
            break;
        case ACCEPT_ACC:
            AcceptISM303_INT_AccExtremes ();
            break;
        case DEFAULTS:
            AcceptISM303_Defaults ();
            break;
        default:
            break;
    }
    int i=0;
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_mag_x_minus100pcReading;
    resp[i++] = ISM303_INT.extremes.minimum.mag.x / 16;
    resp[i++] = ISM303_INT.filtered_mag.x / 16;
    resp[i++] = ISM303_INT.extremes.maximum.mag.x / 16;
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_mag_x_plus100pcReading;

    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_mag_y_minus100pcReading;
    resp[i++] = ISM303_INT.extremes.minimum.mag.y / 16;
    resp[i++] = ISM303_INT.filtered_mag.y / 16;
    resp[i++] = ISM303_INT.extremes.maximum.mag.y / 16;
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_mag_y_plus100pcReading;

    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_mag_z_minus100pcReading;
    resp[i++] = ISM303_INT.extremes.minimum.mag.z / 16;
    resp[i++] = ISM303_INT.filtered_mag.z / 16;
    resp[i++] = ISM303_INT.extremes.maximum.mag.z / 16;
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_mag_z_plus100pcReading;
  
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_acc_x_minus100pcReading;
    resp[i++] = ISM303_INT.extremes.minimum.acc.x / 16;
    resp[i++] = ISM303_INT.filtered_acc.x / 16;
    resp[i++] = ISM303_INT.extremes.maximum.acc.x / 16;
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_acc_x_plus100pcReading;
    
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_acc_y_minus100pcReading;
    resp[i++] = ISM303_INT.extremes.minimum.acc.y / 16;
    resp[i++] = ISM303_INT.filtered_acc.y / 16;
    resp[i++] = ISM303_INT.extremes.maximum.acc.y / 16;
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_acc_y_plus100pcReading;
    
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_acc_z_minus100pcReading;
    resp[i++] = ISM303_INT.extremes.minimum.acc.z / 16;
    resp[i++] = ISM303_INT.filtered_acc.z / 16;
    resp[i++] = ISM303_INT.extremes.maximum.acc.z / 16;
    resp[i++] = OurSettings.SettingsBuffer.ISM303_INT_acc_z_plus100pcReading;

    resp[i++] = (int)(16384 * toolface_m / 360);
    resp[i++] = (int)(16384 * toolface_g / 360);
    resp[i++] = (int)(16384 * toolface_auto / 360);
    resp[i++] = (int)(16384 * inclination / 360);
    

    return i;
}

int prev_rm = -1;
int prev_rt = -1;
int prev_rj = -1;
int prev_rk = -1;
enum {CTRL_MODE_OFF, CTRL_MODE_Position, CTRL_MODE_Speed, CTRL_MODE_AutoSpeedPos, CTRL_MODE_AutoSpeedPos_2, CTRL_MODE_GeoStationary, CTRL_MODE_AntiGeoStationary};

void add_string_to_buffer_a_for_SDcard_file_writes (char *s)
{
    SystemStatusFlags |= flag_in_add_string_to_buffer_a_for_SDcard_file_writes;

    if (mutex)
        SystemStatusFlags |= flag_Buffer_a_over_writing;
    else
    {
        mutex = 1;

        Serial.print(s);

        if (new_ijkmnt)
        {
            new_ijkmnt = 0;

            if (rj != prev_rj)
            {
                prev_rj = rj;
                setpoint_diverter_orientation = (float)(rj & 0x3ff) * 360.0 / 1024;
                Position_Control = 1;
            }
            
            setpoint_RPM = NREF * ((float)ri) / 16384;
            setpoint_RPMfast = NREF * ((float)rn) / 16384;
 
            if (rt != prev_rt)
            {
                prev_rt = rt;
                Timer_TautoTrigger = 0;
                TautoTrigger = (1000 * rt) / 16384;
            }

            if (rk != prev_rk)
            {
                prev_rk = rk;
                ToolFace50t_TriggerOffsetSetting = (float)(rk & 0xfff) * 360.0 / 4096;
            }

            
            if (rm != prev_rm)
            {
                prev_rm = rm;
                control_and_trigger_mode = rm & 0xff;
                                      
                switch (control_and_trigger_mode)
                {
                    case CTRL_MODE_OFF:
                        Position_Control = 0;
                        Speed_Control = 0;
                        digitalWrite(motorEnablePin, LOW);
                        break;
                    case CTRL_MODE_Position:
                    case CTRL_MODE_GeoStationary:
                    case CTRL_MODE_AntiGeoStationary:
                        Position_Control = 1;
                        Speed_Control = 0;
                        break;
                    case CTRL_MODE_Speed:
                        Position_Control = 0;
                        Speed_Control = 1;
                        break;
                    case CTRL_MODE_AutoSpeedPos:
                    case CTRL_MODE_AutoSpeedPos_2:
                        Position_Control = 1;
                        Speed_Control = 0;
                        break;
                  }

                switch (rm >> 8)
                {
                    case 0:
                        SourceForBitOrientation = INCODER_SSI;
                        break;
                    case 1:
                        SourceForBitOrientation = ENCODER_ABZ;
                        break;
                    case 2:
                        SourceForBitOrientation = ISM303_MAGNETIC_TOOLFACE;
                        break;
                    case 3:
                        SourceForBitOrientation = ISM303_GRAVITATIONAL_TOOLFACE;
                        break;
                    case 4:
                        SourceForBitOrientation = ISM303_AUTO_TOOLFACE;
                        break;
                    case 5:
                        SourceForBitOrientation = SIMULATE;
                        break;
                }
            }
        }
        
        while (*s)
        {
            if (Buffer_a_used < BUF_SIZE)
            {
                Buffer_a[Buffer_a_write_pointer++] = *s++;
                Buffer_a_write_pointer &= BUF_SIZE_MASK;
                Buffer_a_used++;
            }
            else
            {
                SystemStatusFlags |= flag_Buffer_a_overflow;
                s++;
            }
        }
        mutex = 0;
    }
    SystemStatusFlags &= ~flag_in_add_string_to_buffer_a_for_SDcard_file_writes;
}

void add_string_to_buffer_b_for_SDcard_file_writes (char *s)
{        
    while (*s)
    {
        if (Buffer_b_used < BUF_SIZE)
        {
            Buffer_b[Buffer_b_write_pointer++] = *s++;
            Buffer_b_write_pointer &= BUF_SIZE_MASK;
            Buffer_b_used++;
        }
        else
        {
            SystemStatusFlags |= flag_Buffer_b_overflow;
            s++;
        }
    }
}

void GetTimeDateNowAsString (char *st)
{
    TimeElements tm;
      
    time_t timenow = Teensy3Clock.get();
    breakTime(timenow, tm);
    sprintf (st, "%02d/%02d/%04d %02d:%02d.%02d", tm.Day, tm.Month, 1970+tm.Year, tm.Hour, tm.Minute, tm.Second);
}

void dateTime(uint16_t* date, uint16_t* time) 
{
    uint16_t year;
    uint8_t month, day, hour, minute, second;
   
    TimeElements tm;
      
    time_t timenow = Teensy3Clock.get();
    breakTime(timenow, tm);
    day = tm.Day;
    month = tm.Month;
    year = 1970+tm.Year;
    hour = tm.Hour;
    minute = tm.Minute;
    second = tm.Second;

    *date = FS_DATE(year, month, day);
    *time = FS_TIME(hour, minute, second);
}

void SD_CARD_cidDmp() 
{
  cid_t cid;
  char st[100];
  uint8_t i;
  
  if (!sd.card()->readCID(&cid)) 
  {
      Serial.println("readCID failed");
  }
  else
  {
      sprintf (st, "SD card Manufacturer ID: 0x%02x OEM ID:0x%02x%02x (%c%c)", (unsigned int)cid.mid, (unsigned int)cid.oid[0], (unsigned int)cid.oid[1],(char)cid.oid[0], (char)cid.oid[1]);
      Serial.println(st);
      Serial.print("Product: ");      
      for (i = 0; i < 5; i++) 
          st[i] = (char)cid.pnm[i];
      st[i]=0;
      Serial.println(st);
      sprintf (st, "Version: %d.%d", (int)cid.prv_n, (int)cid.prv_m);
      Serial.println(st);
      sprintf (st, "Serial number: 0x%08x = %u", (unsigned int)cid.psn, (unsigned int)cid.psn);
      Serial.println(st);
      sprintf (st, "Manufacturing date: %d/%d", (int)cid.mdt_month, 2000 + cid.mdt_year_low + 10 * cid.mdt_year_high);
      Serial.println(st);
  }
}

int prepareSDcard (const char *st_c)
{  
    SystemStatusFlags |= flag_in_prepareSDcard;
    char st[100] = "";

    SdFile::dateTimeCallback(dateTime);
   
    if (sd.begin(SdioConfig(FIFO_SDIO))) 
    {
        SD_CARD_cidDmp();
      
        Serial.print ("SD card formatted for ");  
        sd.printFatType (&Serial);
        
        txtFile = sd.open("boothist.txt", FILE_WRITE);
        GetTimeDateNowAsString (st);
         
        txtFile.print("\n");
        txtFile.print(st_c);
        txtFile.print(" at ");
        txtFile.println(st);
        txtFile.close();
        SystemStatusFlags &= ~flag_in_prepareSDcard;
        sd.end();
        return 1;
    }

    SystemStatusFlags &= ~flag_in_prepareSDcard;
    return 0;
}

void showreg (const char *st, uint32_t r)
{
    Serial.printf("reg %s is 0x%08x\n", st, r);
}

void show_interrupt_priorities ()
{
    Serial.print("Prio IRQ_LPI2C1 is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_LPI2C1));
    Serial.print("Prio IRQ_LPSPI1 is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_LPSPI1));
    Serial.print("Prio IRQ_LPSPI4 is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_LPSPI4));
    Serial.print("Prio IRQ_ALERT_MOTOR is ");
    Serial.println(NVIC_GET_PRIORITY(digitalPinToInterrupt(ina260_motor_alertPin)));
    Serial.print("Prio IRQ_ALERT_COIL is ");
    Serial.println(NVIC_GET_PRIORITY(digitalPinToInterrupt(ina260_coil_alertPin)));
    Serial.print("Prio IRQ_SDHC1 is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_SDHC1));
    Serial.print("Prio IRQ_SDHC2 is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_SDHC2));
    Serial.print("Prio IRQ_USB1 is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_USB1));
    Serial.print("Prio IRQ_USB2 is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_USB2));

    Serial.print("Prio LPUART6 (serial1) is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_LPUART6));
    Serial.print("Prio LPUART4 (serial2) is ");
    Serial.println(NVIC_GET_PRIORITY(IRQ_LPUART4));
}

void setup() 
{
    digitalWrite(PENABLE_Pin, HIGH);
    pinMode(PENABLE_Pin, OUTPUT);
    
    pinMode(ledPin, OUTPUT);

    digitalWrite(OE_ST2378_Pin, LOW);
    pinMode(OE_ST2378_Pin, OUTPUT);
    
    pinMode(ina260_motor_alertPin, INPUT_PULLUP);
    pinMode(ina260_coil_alertPin, INPUT_PULLUP);
    
    pinMode(H_BRIDGE_1_Pin, OUTPUT);
    pinMode(H_BRIDGE_2_Pin, OUTPUT);
    digitalWrite(H_BRIDGE_1_Pin, HIGH);
    digitalWrite(H_BRIDGE_2_Pin, LOW);
//    digitalWrite(H_BRIDGE_2_Pin, HIGH);
//    digitalWrite(H_BRIDGE_1_Pin, LOW);

    setSyncProvider(getTeensy3Time);

    // USB serial port
    Serial.begin(115200);
    int retry = 0;    // Serial is the USB port, but if no PC connected/active, then proceed anyway
    while ((!Serial) && (retry++ <100))
        delay(10);
//    Serial.addMemoryForWrite(SerialWriteBuffer, 1024/2);
//    Serial.addMemoryForWrite(SerialReadBuffer, 1024/2);

    Serial.println("INA260 plus two ST ISM303 3D accel+magnetic sensors plus Maxon Motor test");
    Serial.printf ("Version: %s\nLogFileID:%s\n", versionString, logFileTypeString);
 
    Serial.println ("Starting Serial1 RS485 main");
    pinMode(RS485_DIR1_Pin, OUTPUT);
    digitalWrite(RS485_DIR1_Pin, LOW);  // rx mode, RS485 transmitter off
    Serial1.transmitterEnable (RS485_DIR1_Pin);
    // RS485 main serial port
//    Serial1.begin(115200);
    Serial1.begin(1500000);
    while (!Serial1) 
        delay(10);
        
    Serial1.addMemoryForWrite(Serial1WriteBuffer, 4096/2);
    Serial1.addMemoryForWrite(Serial1ReadBuffer, 4096/2);
    
    Serial1.print("This is RS485 port main at 115200 baud\n");
    Serial1.printf ("Version: %s\nLogFileID:%s\n", versionString, logFileTypeString);
   
    pinMode(RS485_DIR2_Pin, OUTPUT);
    digitalWrite(RS485_DIR2_Pin, LOW);  // rx mode, RS485 transmitter off

      
    Serial.println ("Starting Serial2 RS485 aux");
    // RS485 aux serial port (shared with encoder index inputs)
    Serial2.transmitterEnable (RS485_DIR2_Pin);
    Serial2.begin(1500000);
    while (!Serial2) 
        delay(10);

    pinMode (BinaryInputPremexReedSwitch_Pin, INPUT);

    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_12", IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_12);
    IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_12 |= 0x00010000; // HYSteresis on
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_12", IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_12);
    

    Serial.println ("reading EEPROM setings");
//    OurSettings.show_all_EEPROM_settings_bytes ();
    if (OurSettings.Read_EEPROM_Settings ())
        Serial.println ("read OK");
    else
        Serial.println ("invalid data in EEPROM - switching to defaults");
        
    set_offsets_scales (&ISM303_INT);

    Serial.println ("Starting I2C");
    I2C_master.begin(400 * 1000U);

    Serial.println ("Starting MCP4728 DAC");
    if (!DAC_MCP4728.begin (I2C_master, DAC_MCP4728.I2C_ADDR, LDAC_Pin))
        Serial.println ("Error: MCP4728 DAC registers not read");
    else
    {
        Serial.println ("MCP4728 DAC registers read OK");
//        printDacStatus();
        DAC_MCP4728.selectVref(MCP4728::VREF::VDD, MCP4728::VREF::INTERNAL_2_8V, MCP4728::VREF::INTERNAL_2_8V, MCP4728::VREF::INTERNAL_2_8V, false);
        DAC_MCP4728.selectPowerDown(MCP4728::PWR_DOWN::GND_100KOHM, MCP4728::PWR_DOWN::GND_100KOHM, MCP4728::PWR_DOWN::GND_100KOHM, MCP4728::PWR_DOWN::GND_100KOHM, false);
        DAC_MCP4728.selectGain(MCP4728::GAIN::X1, MCP4728::GAIN::X2, MCP4728::GAIN::X2, MCP4728::GAIN::X2, false);
        DAC_MCP4728.analogWrite(global_DAC_CH_A_value, global_DAC_CH_B_value, global_DAC_CH_C_value, global_DAC_CH_D_value, true);
        
        DAC_MCP4728.enable(true);
    
        DAC_MCP4728.readRegisters();
    }
  
    if (!ina260_motor.begin(I2C_master, INA260_I2CADDR_DEFAULT+1)) 
        Serial.println("Couldn't find INA260 motor chip");
    else
    {
        Serial.println ("Found INA260 motor chip");
   
        ina260_motor.setCurrentConversionTime (INA260_TIME_558_us);
        ina260_motor.setVoltageConversionTime (INA260_TIME_558_us);
        ina260_motor.setAveragingCount (INA260_COUNT_4);   
        ina260_motor.setAlertPolarity (INA260_ALERT_POLARITY_NORMAL);

        ina260_motor.setAlertType (INA260_ALERT_CONVERSION_READY);
        ina260_motor.setMode(INA260_MODE_TRIGGERED);   // trigger first conversion
    }
       
    if (!ina260_coil.begin(I2C_master, INA260_I2CADDR_DEFAULT)) 
        Serial.println("Couldn't find INA260 coil chip");
    else
    {
        Serial.println ("Found INA260 coil chip");
          
        ina260_coil.setCurrentConversionTime (INA260_TIME_558_us);
        ina260_coil.setVoltageConversionTime (INA260_TIME_558_us);
        ina260_coil.setAveragingCount (INA260_COUNT_4);   
        ina260_coil.setAlertPolarity (INA260_ALERT_POLARITY_NORMAL);

        ina260_coil.setAlertType (INA260_ALERT_CONVERSION_READY);
        ina260_coil.setMode(INA260_MODE_TRIGGERED);   // trigger first conversion
    }

    NVIC_SET_PRIORITY(IRQ_LPUART6, 2*16);
    NVIC_SET_PRIORITY(IRQ_LPUART4, 3*16);
 
    NVIC_SET_PRIORITY(IRQ_LPI2C1, 2*16);
    NVIC_SET_PRIORITY(digitalPinToInterrupt(ina260_motor_alertPin), 4*16);
    NVIC_SET_PRIORITY(digitalPinToInterrupt(ina260_coil_alertPin), 4*16);
    
    delay(10); // make sure interrupts from ina260 sensors have triggered already    
    if (ina260_motor.isOK)
        attachInterrupt(digitalPinToInterrupt(ina260_motor_alertPin), ina260_motor_alertPinAction, FALLING);
    if (ina260_coil.isOK)
        attachInterrupt(digitalPinToInterrupt(ina260_coil_alertPin), ina260_coil_alertPinAction, FALLING);
  
/*    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_01", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_01);
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_02", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_02);
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_03", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_03);
*/   
    Serial.println ("Starting SPI");
    SPI.begin();
/*    uint32_t b = 0x8038;
IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_01 = b;
IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_02 = b;
IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_03 = b;
    
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_01", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_01);
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_02", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_02);
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_03", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_03);
*/
    NVIC_SET_PRIORITY(IRQ_LPSPI4, 32);

    int useFIFO = 1;
    int Boot = 1;
    int Magnetic_AC = modulate_AC_on;
    int get_magnetic_vector_fast = 1;
    ISM303_A.init_chip_show_results (useFIFO, Boot, Magnetic_AC, get_magnetic_vector_fast, 0);
    ISM303_B.init_chip_show_results (useFIFO, Boot, Magnetic_AC, get_magnetic_vector_fast, 0);
    ISM303_C.init_chip_show_results (useFIFO, Boot, Magnetic_AC, get_magnetic_vector_fast, 0);
    ISM303_D.init_chip_show_results (useFIFO, Boot, Magnetic_AC, get_magnetic_vector_fast, 0);
    ISM303_INT.init_chip_show_results (useFIFO, Boot, 0, 0, 1);
/*
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_01", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_01);
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_02", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_02);
    showreg ("IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_03", IOMUXC_SW_PAD_CTL_PAD_GPIO_B0_03);
*/

    MaxonMotorEncoderOnTeensy4.setInitConfig();
    MaxonMotorEncoderOnTeensy4.init();

    adc->adc0->setAveraging(16);                                    // set number of averages
    adc->adc0->setResolution(12);                                   // set bits of resolution
    adc->adc0->setConversionSpeed(ADC_CONVERSION_SPEED::MED_SPEED); // change the conversion speed
    adc->adc0->setSamplingSpeed(ADC_SAMPLING_SPEED::MED_SPEED);     // change the sampling speed

    adc->adc1->setAveraging(16);                                    // set number of averages
    adc->adc1->setResolution(12);                                   // set bits of resolution
    adc->adc1->setConversionSpeed(ADC_CONVERSION_SPEED::MED_SPEED); // change the conversion speed
    adc->adc1->setSamplingSpeed(ADC_SAMPLING_SPEED::MED_SPEED);     // change the sampling speed

    pinMode (motorEnablePin, OUTPUT);
    digitalWrite(motorEnablePin, LOW);
    pinMode (motorDirReversePin, OUTPUT);
    digitalWrite(motorDirReversePin, default_forward_def_motor);

    if (prepareSDcard ("Booting"))
        Serial.println ("SD: ok");
    else
        Serial.println ("SD: no compatible microSD card");

    delay (200);

    Serial.println ("Starting orientation controller");
    setup_PI_controller (&PositionController, PosCntrl_P_gain, PosCntrl_I_gain, min_PWM_demanded, max_PWM_demanded);
    setup_PI_controller (&SpeedController, SpeedCntrl_P_gain, SpeedCntrl_I_gain, min_torque_demand, max_torque_demand);
   
    Serial.println ("Starting 10ms and 100ms timers");
      
    myTimer_10ms.begin(timerAction_10ms_prio_2, 10000);  // timerAction to run every 0.01 seconds
    myTimer_100ms.begin(timerAction_100ms_prio_1, 100000);  // timerAction to run every 0.1 seconds

    myTimer_10ms.priority (3*16); //  lower number = higher priority
    myTimer_100ms.priority (5*16); //  lower than 96 number = low priority, but higher than SD card

    NVIC_SET_PRIORITY(IRQ_SDHC1, 11*16);

    show_interrupt_priorities ();
}

void printDacStatus()
{
    Serial.println("NAME     Vref  Gain  PowerDown  DACData");
    for (int i = 0; i < 4; ++i)
    {
        Serial.print("DAC");
        Serial.print(i, DEC);
        Serial.print("   ");
        Serial.print("    ");
        Serial.print(DAC_MCP4728.getVref(i), BIN);
        Serial.print("     ");
        Serial.print(DAC_MCP4728.getGain(i), BIN);
        Serial.print("       ");
        Serial.print(DAC_MCP4728.getPowerDown(i), BIN);
        Serial.print("       ");
        Serial.println(DAC_MCP4728.getDACData(i), DEC);

        Serial.print("EEPROM");
        Serial.print(i, DEC);
        Serial.print("    ");
        Serial.print(DAC_MCP4728.getVref(i, true), BIN);
        Serial.print("     ");
        Serial.print(DAC_MCP4728.getGain(i, true), BIN);
        Serial.print("       ");
        Serial.print(DAC_MCP4728.getPowerDown(i, true), BIN);
        Serial.print("       ");
        Serial.println(DAC_MCP4728.getDACData(i, true), DEC);
    }
    Serial.println(" ");
}

void ina260_motor_alertPinAction ()
{
    SystemStatusFlags |= flag_in_isr_alertPinAction;

    float current, voltage, power;

    ina260_motor.readCurrVoltPower (&current, &voltage, &power, INA260_MODE_TRIGGERED);   // trigger next conversion

    motor_current = current;

    if (flag_clear_current_averages_motor)
    {
        current_integrator_motor = current;
        voltage_integrator_motor = voltage;
        power_integrator_motor = power;
        nav_current_motor = 2;
        flag_clear_current_averages_motor = 0;
    }
    else
    {
        nav_current_motor++;
        current_integrator_motor += current;
        voltage_integrator_motor += voltage;
        power_integrator_motor += power;
        nav_current_motor++;
    }

    Count_10ms_ticks_no_Alert_isr_motor = 0;
    Count_100ms_ticks_no_Alert_isr_motor = 0;

    if (flag_new_dac_values)
    {
        DAC_MCP4728.analogWrite(global_DAC_CH_A_value, global_DAC_CH_B_value, global_DAC_CH_C_value, global_DAC_CH_D_value, false);
        flag_new_dac_values = 0;
    }

    SystemStatusFlags &= ~flag_in_isr_alertPinAction;
}

void ina260_coil_alertPinAction ()
{
    SystemStatusFlags |= flag_in_isr_alertPinAction;

    float current, voltage, power;
 
    ina260_coil.readCurrVoltPower (&current, &voltage, &power, INA260_MODE_TRIGGERED);   // trigger next conversion
    
    if (flag_clear_current_averages_coil)
    {
        current_integrator_coil = current;
        voltage_integrator_coil = voltage;
        power_integrator_coil = power;
        nav_current_coil = 2;
        flag_clear_current_averages_coil = 0;
    }
    else
    {
        nav_current_coil++;
        current_integrator_coil += current;
        voltage_integrator_coil += voltage;
        power_integrator_coil += power;
        nav_current_coil++;
    }

    Count_10ms_ticks_no_Alert_isr_coil = 0;
    Count_100ms_ticks_no_Alert_isr_coil = 0;
    SystemStatusFlags &= ~flag_in_isr_alertPinAction;
}

int cap0to4095_scale (float scale, float value)
{
    float r = scale * value;
    if (r<0.0)
        r = 0.0;
    if (r > 4095.0)
        r = 4095.0;
  
    return (int)r;
}

int quadrant (float alfa)
{
  return (((int)alfa )/90);
}

float delta_angle (float a, float b)
{
  int quadrant_a = quadrant (a);
  int quadrant_b = quadrant (b);
  float delta = a - b;
  switch (quadrant_a)
  {
    case 0:
            switch (quadrant_b)
            {
                case 0: return delta;
                case 1: return delta;
                case 2: if (delta < -180.0)
                            return 360.0 + delta;
                        else
                            return delta;
                case 3: return 360.0 + delta;
            }
    case 1:
            switch (quadrant_b)
            {
                case 0: return delta;
                case 1: return delta;
                case 2: return delta;
                case 3: if (delta < -180.0)
                            return 360.0 + delta;
                        else
                            return delta;
            }
    case 2:
            switch (quadrant_b)
            {
                case 0: if (delta > 180.0)
                            return delta - 360;
                        else
                            return delta;
                case 1: return delta;
                case 2: return delta;
                case 3: return delta;
            }
     case 3:
            switch (quadrant_b)
            {
                case 0: return delta - 360;
                case 1: if (delta > 180.0)
                            return delta - 360;
                        else
                            return delta;
                case 2: return delta;
                case 3: return delta;
            }
    }
    return 0;
}

float MakeSureItIsFrom0to360 (float x)
{
    while (x < 0.0)
        x += 360.0;
    while (x >= 360.0)
        x -= 360.0;
    return x;
}

volatile float past5samples[5] = {0.0, 0.0, 0.0, 0.0, 0.0};

float max_in_past_5_samples (float new_x)
{
   int i;
   int N = 5;
   float maximum;
   
   for (i=N-1; i>0; i--)
      past5samples[i] = past5samples[i-1];
 
   past5samples[0] = new_x;

   maximum = 0.0;
   for (i=0; i<N; i++)
       if (past5samples[i] > maximum)
          maximum = past5samples[i];
 
   return maximum;
}

void set_PWM_demanded (int x)
{
    PWM_demanded = x;
    if (PWM_demanded > 0)
    {
        digitalWrite(motorDirReversePin, default_forward_def_motor);
        global_DAC_CH_A_value = PWM_demanded;
    }
    else
    {
        digitalWrite(motorDirReversePin, !default_forward_def_motor);
        global_DAC_CH_A_value = -PWM_demanded;
    }
    flag_new_dac_values = 1;
    digitalWrite(motorEnablePin, HIGH);
}

// 2^10 = 1024 = 10.24 s max delay at 100 s/s
#define FIFOsize (1<<10)
#define FIFOmask (FIFOsize-1)
volatile double FIFO_delay_line[FIFOsize];
volatile int FIFO_pos_write = 0;

double FIFO_push_pull (double new_input, uint32_t NumberOfSamplesDelay)
{
    if (NumberOfSamplesDelay >= FIFOsize)
        NumberOfSamplesDelay = FIFOsize-1;
    int FIFO_pos_read = (FIFO_pos_write - (NumberOfSamplesDelay-1)) & FIFOmask;
    double r = FIFO_delay_line[FIFO_pos_read];
    FIFO_delay_line [FIFO_pos_write] = new_input;
    FIFO_pos_write++;
    FIFO_pos_write &= FIFOmask;
    return r;
}


void sense_BinaryInputPremexReedSwitch () 
{
    if (digitalRead (BinaryInputPremexReedSwitch_Pin))
    {
        if (IntInputPremexReedSwitch < 10)
            IntInputPremexReedSwitch++;
    }
    else
    {
        if (IntInputPremexReedSwitch > 0)
            IntInputPremexReedSwitch--;
    }
}

void timerAction_10ms_prio_2 () 
{
    char st10[256] = "";
    Three_D_vector_t AccelSensors3D;
    Three_D_vector_t MagneticSensors3D;
    
    sense_BinaryInputPremexReedSwitch ();
    
    pos_encoder = MaxonMotorEncoderOnTeensy4.read() >> 2;
    
    sense_BinaryInputPremexReedSwitch ();
    
    speed_encoder = pos_encoder - previous_pos_encoder;
    previous_pos_encoder = pos_encoder;

    pos_encoder_int32_sync += speed_encoder;

    if (pos_encoder_int32_sync >= ShaftSyncTurn)
        pos_encoder_int32_sync -= ShaftSyncTurn;

    if (pos_encoder_int32_sync <= -ShaftSyncTurn)
        pos_encoder_int32_sync += ShaftSyncTurn;  

    if (flag_clear_encoder_averages)
    {
        pos_encoder_av = pos_encoder;
        speed_encoder_av = speed_encoder;
        encoder_nav = 2;
        flag_clear_encoder_averages = 0;
    }
    else
    {
        encoder_nav++;
        pos_encoder_av += pos_encoder;
        speed_encoder_av += speed_encoder;
        encoder_nav++;
    }
    
    ISM303_INT.getAllSensorValues (toggle_bit);
    sense_BinaryInputPremexReedSwitch ();
    ISM303_A.getAllSensorValues (toggle_bit);
    sense_BinaryInputPremexReedSwitch ();
    ISM303_B.getAllSensorValues (toggle_bit);
    sense_BinaryInputPremexReedSwitch ();
    ISM303_C.getAllSensorValues (toggle_bit);
    sense_BinaryInputPremexReedSwitch ();
    ISM303_D.getAllSensorValues (toggle_bit);

    global_DAC_CH_B_value = cap0to4095_scale (0.5, ISM303_A.mag_magnitude_AC_fast);
    global_DAC_CH_C_value = cap0to4095_scale (0.5, ISM303_B.mag_magnitude_AC_fast);
    global_DAC_CH_D_value = cap0to4095_scale (0.5, ISM303_C.mag_magnitude_AC_fast);
    flag_new_dac_values = 1;    

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE) 
    {
        if (toggle_bit)
        {
            digitalWrite(H_BRIDGE_1_Pin, HIGH);
            digitalWrite(H_BRIDGE_2_Pin, LOW);
            toggle_bit = 0;
        }
        else
        {
            if (modulate_AC_on)
            {
                digitalWrite(H_BRIDGE_2_Pin, HIGH);
                digitalWrite(H_BRIDGE_1_Pin, LOW);
            }
            toggle_bit = 1;
        }
    }

    exit_shaft_orientation = 360.0 * ((double)(pos_encoder_int32_sync % EncoderPulsesPerDiverterRev) / EncoderPulsesPerDiverterRev);
    
    RPM_exit_shaft = 60 * 100 * (double)speed_encoder * GearBoxRatioD / (EncoderPulsesPerMotorRev * GearBoxRatioN);
 
    sense_BinaryInputPremexReedSwitch ();

    FilteredBinaryInputPremexReedSwitch = (IntInputPremexReedSwitch > 5);

    if (FilteredBinaryInputPremexReedSwitch)
    {
        if (!duration_PremexReedSwitch_high)
        {
            prev_duration_PremexReedSwitch_low = duration_PremexReedSwitch_low;
        }
        duration_PremexReedSwitch_high++;   // debounce switch contact
        duration_PremexReedSwitch_low = 0;
    }
    else
    {
        if (!duration_PremexReedSwitch_low)
        {
            prev_duration_PremexReedSwitch_high = duration_PremexReedSwitch_high;
        }
        duration_PremexReedSwitch_low++;
        duration_PremexReedSwitch_high = 0;
    }
    
    if (RPM_exit_shaft > 0)
    {
        if (duration_PremexReedSwitch_low == 1)    // falling edge? (=relay switch goes on, pulls io pin low)
           if (prev_duration_PremexReedSwitch_high > 8)
               exit_shaft_orientation_at_Premex_edge = exit_shaft_orientation;
    }
    else
    {
        if (duration_PremexReedSwitch_high == 1)    // rising edge? (=relay switch goes off, pulls io pin high)
           if (prev_duration_PremexReedSwitch_low > 8)
               exit_shaft_orientation_at_Premex_edge = exit_shaft_orientation;
    }

    switch (SourceForBitOrientation)
    {
        case INCODER_SSI:
            send_poll_for_50t_encoder (1);
            break;
        
        case ENCODER_ABZ:
            send_poll_for_50t_encoder (0);
            break;

        case ISM303_MAGNETIC_TOOLFACE:
            Encoder50t = (int)(16383 * toolface_m / 360.0);
            Encoder50t &= 0x3fff;
            EncWordToPC = 0x4000 | Encoder50t;
            break;
 
        case ISM303_GRAVITATIONAL_TOOLFACE:
            Encoder50t = (int)(16383 * toolface_g / 360.0);
            Encoder50t &= 0x3fff;
            EncWordToPC = 0x4000 | Encoder50t;
            
            break;
        case ISM303_AUTO_TOOLFACE:
            Encoder50t = (int)(16383 * toolface_auto / 360.0);
            Encoder50t &= 0x3fff;
            EncWordToPC = 0x4000 | Encoder50t;    
            break;
            
        case SIMULATE:
            Encoder50t += 40;
            Encoder50t &= 0x3fff;
            EncWordToPC = 0x4000 | Encoder50t;
            break;
    }
 
    ShaftEncoderAndPremexReedSwitch = EncWordToPC & 0x7fff;

    if (FilteredBinaryInputPremexReedSwitch)
        ShaftEncoderAndPremexReedSwitch |= 0x8000;

    if (EncWordToPC & 0x4000)
        ToolFace50t_Encoder =  (float)(Encoder50t & 0x3fff) * 360.0 / 16384; 
        
    Shifted50t_encoder = MakeSureItIsFrom0to360 (ToolFace50t_Encoder - ToolFace50t_TriggerOffsetSetting + 180.0);

    QuadrantWeReIn = (int)(Shifted50t_encoder / 90.0);
    bool do_trigger = 0;
 
    if ((QuadrantWeReIn == 2) && (prev_QuadrantWeReIn == 1) && passed360to0)
    {
        do_trigger = 1;
        passed360to0 = 0;
    }
    if ((QuadrantWeReIn == 0) && (prev_QuadrantWeReIn == 3))
        passed360to0 = 1;
    
    prev_QuadrantWeReIn = QuadrantWeReIn;


    if (syncOnPremex)
        measured_diverter_orientation = MakeSureItIsFrom0to360 (exit_shaft_orientation - exit_shaft_orientation_at_Premex_edge);
    else
        measured_diverter_orientation = exit_shaft_orientation;

    switch (control_and_trigger_mode)
    {
        case CTRL_MODE_Position:  
        case CTRL_MODE_AutoSpeedPos:  
            orientation_error = delta_angle ( MakeSureItIsFrom0to360 (setpoint_diverter_orientation), measured_diverter_orientation);
            break;
        case CTRL_MODE_AutoSpeedPos_2:
            orientation_error = delta_angle ( MakeSureItIsFrom0to360 (setpoint_diverter_orientation + (Shifted50t_encoder - 180.0)), measured_diverter_orientation);
            break;
        case CTRL_MODE_GeoStationary:
            orientation_error = delta_angle ( MakeSureItIsFrom0to360 ((Shifted50t_encoder - 180.0)), measured_diverter_orientation);
            break;
        case CTRL_MODE_AntiGeoStationary:
            orientation_error = delta_angle ( MakeSureItIsFrom0to360 ( - (Shifted50t_encoder - 180.0)), measured_diverter_orientation);
            break;
            
    }

    orientation_step = delta_angle (measured_diverter_orientation, prev_measured_diverter_orientation);
    prev_measured_diverter_orientation = measured_diverter_orientation;

    integrated_orientation_steps += orientation_step;


    if (Position_Control && (!Speed_Control))
    {
        PI_control_step (orientation_error, &PositionController);
    
        set_PWM_demanded ((int)PositionController.PI_out);

        if (control_and_trigger_mode == CTRL_MODE_AutoSpeedPos)
        {
            if (max_in_past_5_samples (fabs(orientation_error)) < 2.0)    // reached target position within +/- 2.0 deg and steady there?
            {
                digitalWrite(motorEnablePin, LOW);
                PWM_demanded = 0;
                Position_Control = 0;   // full rotation job done, wait for next trigger
            }
            else
                digitalWrite(motorEnablePin, HIGH);
        }
        
        setpoint_RPM_exit_shaft = setpoint_RPM; // for when the next 'go' trigger comes

    }
    else
        if (Speed_Control)
        {
            if (!prevSpeed_Control)
            {
                SpeedController.I_contribution = 0.0;
                integrated_orientation_steps = 0;
            }

            if ((control_and_trigger_mode != CTRL_MODE_AutoSpeedPos) && (control_and_trigger_mode != CTRL_MODE_AutoSpeedPos_2))
                setpoint_RPM_exit_shaft = setpoint_RPM;
            else
            {
         
 //               if (integrated_orientation_steps < 90.0)    // first 90 degrees: travel with well defined setpoint speed
                if (integrated_orientation_steps < 180.0)    // first 180 degrees: travel with well defined setpoint speed
                    setpoint_RPM_exit_shaft = setpoint_RPM;
                else
                {
                    switch (control_and_trigger_mode)
                    {
                        case CTRL_MODE_AutoSpeedPos:
                            if (integrated_orientation_steps < 240.0)    // up to 240 degrees: travel with very high speed, only forward
                                setpoint_RPM_exit_shaft = setpoint_RPMfast;
                            else
                            {
                                Speed_Control = 0;
                                Position_Control = 1;   // travelled through window open period, now go to park position
                            }
                            break;
                        case CTRL_MODE_AutoSpeedPos_2:
                            if (orientation_error < 0)
                                setpoint_RPM_exit_shaft = 0;
                            else
                            {
                                Speed_Control = 0;
                                Position_Control = 1;   // travelled through window open period, now go to park position while latching onto 50t orientation
                            }
                            break;
                    }
                    
                }
            }    
            
            PositionController.I_contribution = 0.0;
    
            speed_error = setpoint_RPM_exit_shaft - RPM_exit_shaft;
           
            if (Speed_Control)
            {
                PI_control_step (speed_error, &SpeedController);
                
                set_PWM_demanded ((int)SpeedController.PI_out);
            }
            
        }
        else      // not position and not speed control => turn off Maxon PWM controller
        {
            set_PWM_demanded (0);
            digitalWrite(motorEnablePin, LOW);
        }
    prevSpeed_Control = Speed_Control;

    if (TautoTrigger)   // use internal free running timer or actual 50t encoder as trigger?
    {
        Timer_TautoTrigger++;
        if (Timer_TautoTrigger >= TautoTrigger)
        {
            Timer_TautoTrigger = 0;
            do_trigger = 1;
        }
    }

    if (do_trigger)
        if (!Speed_Control)
            switch (control_and_trigger_mode)
            {
                case CTRL_MODE_AutoSpeedPos:
                case CTRL_MODE_AutoSpeedPos_2:
                    Speed_Control = 1;
                    break;
            }

    
    if (n_10ms_samples_since_boot++)
    {
        MA = ISM303_A.mag_magnitude_AC_fast;
        MB = ISM303_B.mag_magnitude_AC_fast;
        MC = ISM303_C.mag_magnitude_AC_fast;
        MD = ISM303_D.mag_magnitude_AC_fast;
        
        ME = FIFO_push_pull (MD, FIFO_delay_D_to_Bit);
 sense_BinaryInputPremexReedSwitch ();
       
        sprintf (st10, "%s %ld ", logFileTypeString, n_10ms_samples_since_boot);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
       
        GetTimeDateNowAsString (st10);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
sense_BinaryInputPremexReedSwitch ();
        
        sprintf (st10, " %1.1f %1.1f %1.1f %1.1f %1.1f", MA, MB, MC, MD, ME);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
sense_BinaryInputPremexReedSwitch ();

        sprintf (st10, " %d" , Encoder50t);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
sense_BinaryInputPremexReedSwitch ();

        sprintf (st10, " %d %1.1f %1.1f %1.1f", (int)pos_encoder, exit_shaft_orientation, setpoint_RPM_exit_shaft, RPM_exit_shaft);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
sense_BinaryInputPremexReedSwitch ();
        
        sprintf (st10, " %d", (int)FilteredBinaryInputPremexReedSwitch);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
sense_BinaryInputPremexReedSwitch ();

        sprintf (st10, " %d", 2*(int)Position_Control + (int)Speed_Control);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
sense_BinaryInputPremexReedSwitch ();

        sprintf (st10, " %1.1f %1.1f %1.1f %d", setpoint_diverter_orientation, measured_diverter_orientation, orientation_error, PWM_demanded);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
sense_BinaryInputPremexReedSwitch ();

        sprintf (st10, " %d %d %d", (int)Xcorr_Velocity_as_int16, (int)Xcorr_Corr_Coeff_as_int16, (int)Spare3);
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
                
sense_BinaryInputPremexReedSwitch ();
        sprintf (st10, "\n");
        add_string_to_buffer_b_for_SDcard_file_writes (st10);
    }

    push_10ms_values_in_elastic_buffer ();
  
    if (++Count_10ms_ticks >= 100)
    {
        Count_10ms_ticks = 0;
        UpTimeSeconds++;
    }

    
    if (++Count2_10ms_ticks >= BlinkDurationIn10msCounts)
    {
        Count2_10ms_ticks = 0;
   
        if (ledState == LOW)
             ledState = HIGH;
        else 
            ledState = LOW;
        
//        digitalWrite(ledPin, ledState);
    }

    AccelSensors3D.X = (float)ISM303_INT.acc_f.x / ACC_100pc;
    AccelSensors3D.Y = (float)ISM303_INT.acc_f.z / ACC_100pc;
    AccelSensors3D.Z = (float)ISM303_INT.acc_f.y / ACC_100pc;
    
    MagneticSensors3D.X = -(float)ISM303_INT.mag_f.x / MAG_100pc;
    MagneticSensors3D.Y = -(float)ISM303_INT.mag_f.z / MAG_100pc;
    MagneticSensors3D.Z = (float)ISM303_INT.mag_f.y / MAG_100pc;

    normalise_to_one_dot_zero_magnitude (&AccelSensors3D);
    normalise_to_one_dot_zero_magnitude (&MagneticSensors3D);
sense_BinaryInputPremexReedSwitch ();

    ISM303_INT.flag_get_averages = 1;

    populate_TriGonioFloats_from_3Daccel_and_3Dmag (&TriGonioFloats, &AccelSensors3D, &MagneticSensors3D);
sense_BinaryInputPremexReedSwitch ();
            
    inclination = Inclination(&TriGonioFloats);
    total_grav = TotGravFld (&TriGonioFloats);
    total_magn = TotMagFld (&TriGonioFloats);
    toolface_g = GravToolFace(&TriGonioFloats);
    toolface_m = MagneticTF(&TriGonioFloats);    
    azimuth = Azimuth(&TriGonioFloats);
    dipangle = DipAngle(&TriGonioFloats);
sense_BinaryInputPremexReedSwitch ();

    if ((inclination > 170) || (inclination < 10))
        TF_auto_mode = MAGNETIC;
    if ((inclination < 165) || (inclination > 15))
        TF_auto_mode = GRAVITATIONAL;

    if (TF_auto_mode == GRAVITATIONAL)
        toolface_auto = toolface_g;
    else
        toolface_auto = toolface_m;
            
 
    if (++Count_10ms_ticks_no_100ms_isr > 11)   // timeout 100 ms timer interrupt?
        SystemStatusFlags |= flag_missed_100ms_isr;

    if (++Count_10ms_missed_loop_cycles > 100)   // loop code not repeated within 1000 ms?
        SystemStatusFlags |= flag_missed_loop_cycles;

    if (++Count_10ms_missed_SD_write > 1000)   // SD card write not repeated within 10 s?
        SystemStatusFlags |= flag_missed_SD_write;
   
    if (Count_10ms_missed_loop_cycles < 20)   // any loop cycles done in past 200 ms?
        SystemStatusFlags |= flag_did_loop_in_past_200ms;
    else
        SystemStatusFlags &= ~flag_did_loop_in_past_200ms;

    if (Count_10ms_ticks_no_Alert_isr_motor < 20)   // any INA260 serviced in past 200ms ?
        SystemStatusFlags |= flag_did_INA260_motor_in_past_200ms;
    else
        SystemStatusFlags &= ~flag_did_INA260_motor_in_past_200ms;
        
    if (Count_10ms_ticks_no_Alert_isr_coil < 20)   // any INA260 serviced in past 200ms ?
        SystemStatusFlags |= flag_did_INA260_coil_in_past_200ms;
    else
        SystemStatusFlags &= ~flag_did_INA260_coil_in_past_200ms;
        
    if (Count_10ms_missed_SD_write < 100)   // SD card written to OK in past 1000 ms?
        SystemStatusFlags |= flag_did_SD_write_in_past_second;
    else
        SystemStatusFlags &= ~flag_did_SD_write_in_past_second;
}

bool mutex3 = 0;

void timerAction_100ms_prio_1 () 
{
    SystemStatusFlags |= flag_in_isr_100ms;

  
    int nav_current_before, nav_current_after_motor = 0, nav_current_after_coil = 0;
    float current_for_printing_motor = 0.0, voltage_for_printing_motor = 0.0; 
// float power_for_printing_motor = 0.0;
    float current_for_printing_coil = 0.0, voltage_for_printing_coil = 0.0; 
// float power_for_printing_coil = 0.0;
    float copy_current_integrator, copy_voltage_integrator; 
// float copy_power_integrator;

    int32_t copy_pos_encoder_av;
    int32_t copy_speed_encoder_av;
    int32_t copy_encoder_nav = 0, copy_encoder_nav_before;
    
    char st[256] = "";

/*    
  Three_D_vector_t AccelSensors3D;
  Three_D_vector_t MagneticSensors3D;
*/
    if (mutex3)
        SystemStatusFlags |= flag_overasking_100ms_actions;
    else
    {
        if (flag_next_1seconds_trigger)
        {
            flag_1seconds_trigger = 1;
            flag_next_1seconds_trigger = 0;
        }
        if (++Count_100ms_ticks_for_flag_1seconds_trigger > 10)
        {
            flag_1seconds_trigger = 1;
            Count_100ms_ticks_for_flag_1seconds_trigger = 0;
        }
      
        mutex3 = 1;
        if (nav_current_motor > 1)
        {
            do
            {
                nav_current_before = nav_current_motor;
                copy_current_integrator = current_integrator_motor;
                copy_voltage_integrator = voltage_integrator_motor;
//                copy_power_integrator = power_integrator_motor;
                nav_current_after_motor = nav_current_motor;
            } while (nav_current_before != nav_current_after_motor);
            
            flag_clear_current_averages_motor = 1;
            nav_current_after_motor >>= 1;
        
            current_for_printing_motor = copy_current_integrator / nav_current_after_motor;
            voltage_for_printing_motor = copy_voltage_integrator / nav_current_after_motor;
//            power_for_printing_motor = copy_power_integrator / nav_current_after_motor;
        }
 
        if (nav_current_coil > 1)
        {
            do
            {
                nav_current_before = nav_current_coil;
                copy_current_integrator = current_integrator_coil;
                copy_voltage_integrator = voltage_integrator_coil;
//                copy_power_integrator = power_integrator_coil;
                nav_current_after_coil = nav_current_coil;
            } while (nav_current_before != nav_current_after_coil);
            
            flag_clear_current_averages_coil = 1;
            nav_current_after_coil >>= 1;
        
            current_for_printing_coil = copy_current_integrator / nav_current_after_coil;
            voltage_for_printing_coil = copy_voltage_integrator / nav_current_after_coil;
//            power_for_printing_coil = copy_power_integrator / nav_current_after_coil;
        }

        if (encoder_nav > 1)
        {
            do
            {
                copy_encoder_nav_before = encoder_nav;
                copy_pos_encoder_av = pos_encoder_av;
                copy_speed_encoder_av = speed_encoder_av;
                copy_encoder_nav = encoder_nav;
            } while (copy_encoder_nav != copy_encoder_nav_before);
            
            flag_clear_encoder_averages = 1;
            copy_encoder_nav >>= 1;
            
            encoder_pos = copy_pos_encoder_av / copy_encoder_nav;
            encoder_speed = copy_speed_encoder_av / copy_encoder_nav;
        }
        if (n_100ms_samples_since_boot)
        {
            sprintf (st, "%s %d ", logFileTypeString, (int)n_100ms_samples_since_boot);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
           
            GetTimeDateNowAsString (st);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
            
            sprintf (st, " %1.1f %1.1f", current_for_printing_motor, voltage_for_printing_motor);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
            sprintf (st, " %1.1f %1.1f", current_for_printing_coil, voltage_for_printing_coil);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
            
            if (Serial.available()) 
            {
                time_t t = processSyncMessage();
                if (t != 0) 
                {
                    Teensy3Clock.set(t); // set the RTC
                    setTime(t);
                }
            }
        
            ISM303_INT.printAllSensorValues (ACC_XYZ | MAG_XYZ | TEMPERATURE);
            ISM303_A.printAllSensorValues (MAG_MAGNITUDE);
            ISM303_B.printAllSensorValues (MAG_MAGNITUDE);
            ISM303_C.printAllSensorValues (MAG_MAGNITUDE);
            ISM303_D.printAllSensorValues (MAG_MAGNITUDE);

 /*           
            AccelSensors3D.X = (float)ISM303_INT.acc_f.x / ACC_100pc;
            AccelSensors3D.Y = (float)ISM303_INT.acc_f.z / ACC_100pc;
            AccelSensors3D.Z = (float)ISM303_INT.acc_f.y / ACC_100pc;
            
            MagneticSensors3D.X = -(float)ISM303_INT.mag_f.x / MAG_100pc;
            MagneticSensors3D.Y = -(float)ISM303_INT.mag_f.z / MAG_100pc;
            MagneticSensors3D.Z = (float)ISM303_INT.mag_f.y / MAG_100pc;

            normalise_to_one_dot_zero_magnitude (&AccelSensors3D);
            normalise_to_one_dot_zero_magnitude (&MagneticSensors3D);


            ISM303_INT.flag_get_averages = 1;
 */
            ISM303_A.flag_get_averages = 1;
            ISM303_B.flag_get_averages = 1;
            ISM303_C.flag_get_averages = 1;
            ISM303_D.flag_get_averages = 1;

            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_INT.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_A.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_B.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_C.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_D.txt_st);
            
 /*           
            populate_TriGonioFloats_from_3Daccel_and_3Dmag (&TriGonioFloats, &AccelSensors3D, &MagneticSensors3D);
            
            inclination = Inclination(&TriGonioFloats);
            total_grav = TotGravFld (&TriGonioFloats);
            total_magn = TotMagFld (&TriGonioFloats);
            toolface_g = GravToolFace(&TriGonioFloats);
            toolface_m = MagneticTF(&TriGonioFloats);    
            azimuth = Azimuth(&TriGonioFloats);
            dipangle = DipAngle(&TriGonioFloats);

            if ((inclination > 170) || (inclination < 10))
                TF_auto_mode = MAGNETIC;
            if ((inclination < 165) || (inclination > 15))
                TF_auto_mode = GRAVITATIONAL;

            if (TF_auto_mode == GRAVITATIONAL)
                toolface_auto = toolface_g;
            else
                toolface_auto = toolface_m;
 */
            
            if (zero_rt)
            {
                MaxonMotorEncoderOnTeensy4.write(0);
                zero_rt = 0;
            }

       
            sprintf (st, " p=%d" , Encoder50t);
            add_string_to_buffer_a_for_SDcard_file_writes (st);

            sprintf (st, " o=%3.2f rs=%3.2f ra=%3.2f", exit_shaft_orientation, setpoint_RPM_exit_shaft, RPM_exit_shaft);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
            
            sprintf (st, " %d", (int)FilteredBinaryInputPremexReedSwitch);
            add_string_to_buffer_a_for_SDcard_file_writes (st);

            sprintf (st, " %d %1.1f", 2*(int)Position_Control + (int)Speed_Control, setpoint_diverter_orientation);
            add_string_to_buffer_a_for_SDcard_file_writes (st);

            sprintf (st, " %1.1f %1.1f eO=%1.1f eR=%1.0f q=%d", setpoint_diverter_orientation, measured_diverter_orientation, orientation_error, speed_error, PWM_demanded);
            add_string_to_buffer_a_for_SDcard_file_writes (st);

            sprintf (st, " %d %d %d", (int)Xcorr_Velocity_as_int16, (int)Xcorr_Corr_Coeff_as_int16, (int)Spare3);
            add_string_to_buffer_a_for_SDcard_file_writes (st);


            if (MotorTestScript)
                if (++Count_100ms_ticks >= 10)
                {
                    Count_100ms_ticks = 0;
                    motor_test_script_second++;
                  
                    switch (motor_test_script_step)
                    {
                        case 0:
                            digitalWrite(motorEnablePin, LOW);
                            global_DAC_CH_A_value = 0;
                            if (motor_test_script_second >= 2)
                            {
                                motor_test_script_second = 0;
                                motor_test_script_step++;
                            }
                           break;
                        case 1:
                            digitalWrite(motorEnablePin, HIGH);
                            digitalWrite(motorDirReversePin, default_forward_def_motor);
                            global_DAC_CH_A_value = (motor_test_script_second) * 4096 / 11;
                            if (motor_test_script_second >= 10)
                            {
                                motor_test_script_second = 0;
                                motor_test_script_step++;
                            }
                            break;
                        case 2:
                            digitalWrite(motorEnablePin, HIGH);
                            digitalWrite(motorDirReversePin, default_forward_def_motor);
                            global_DAC_CH_A_value = (10 - motor_test_script_second) * 4096 / 11;
       
                            if (motor_test_script_second >= 10)
                            {
                                motor_test_script_second = 0;
                                motor_test_script_step++;
                            }
                            break;
                        case 3:
                            digitalWrite(motorEnablePin, HIGH);
                            digitalWrite(motorDirReversePin, !default_forward_def_motor);
                            global_DAC_CH_A_value = (motor_test_script_second) * 4096 / 11;
                           if (motor_test_script_second >= 10)
                            {
                                motor_test_script_second = 0;
                                motor_test_script_step++;
                            }
                            break;
                        case 4:
                            digitalWrite(motorEnablePin, HIGH);
                            digitalWrite(motorDirReversePin, !default_forward_def_motor);
                            global_DAC_CH_A_value = (10 - motor_test_script_second) * 4096 / 11;
       
                            if (motor_test_script_second >= 10)
                            {
                                motor_test_script_second = 0;
                                motor_test_script_step = 0;
                            }
                            break;
                    }           
                    flag_new_dac_values = 1;
                }
                
 //           int valADC = adc->adc1->analogRead(A17);  

            if (show_magnetic_extremes_ISM303_INT)
            {
                sprintf (st, " MX<%d %d %d> MY<%d %d %d> MZ<%d %d %d>", 
                            ISM303_INT.extremes.minimum.mag.x / 16, ISM303_INT.filtered_mag.x / 16, ISM303_INT.extremes.maximum.mag.x / 16, 
                            ISM303_INT.extremes.minimum.mag.y / 16, ISM303_INT.filtered_mag.y / 16, ISM303_INT.extremes.maximum.mag.y / 16, 
                            ISM303_INT.extremes.minimum.mag.z / 16, ISM303_INT.filtered_mag.z / 16, ISM303_INT.extremes.maximum.mag.z / 16);
                add_string_to_buffer_a_for_SDcard_file_writes (st);
            }
            if (show_acceleration_extremes_ISM303_INT)
            {
                sprintf (st, " GX<%d %d %d> GY<%d %d %d> GZ<%d %d %d>", 
                            ISM303_INT.extremes.minimum.acc.x / 16, ISM303_INT.filtered_acc.x / 16, ISM303_INT.extremes.maximum.acc.x / 16, 
                            ISM303_INT.extremes.minimum.acc.y / 16, ISM303_INT.filtered_acc.y / 16, ISM303_INT.extremes.maximum.acc.y / 16, 
                            ISM303_INT.extremes.minimum.acc.z / 16, ISM303_INT.filtered_acc.z / 16, ISM303_INT.extremes.maximum.acc.z / 16);
                add_string_to_buffer_a_for_SDcard_file_writes (st);
            }

            sprintf (st, "\n");
            add_string_to_buffer_a_for_SDcard_file_writes (st);
  /*         
            sprintf (st, "%s %08x %d %ld %d ", logFileTypeString, (int)SystemStatusFlags, debug, n_100ms_samples_since_boot, BlinkDurationIn10msCounts);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
           
            GetTimeDateNowAsString (st);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
            
            sprintf (st, " M: I=%4.0f, V=%4.0f P=%4.0f (%d)", current_for_printing_motor, voltage_for_printing_motor, power_for_printing_motor, nav_current_after_motor);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
            sprintf (st, " C: I=%4.0f, V=%4.0f P=%4.0f (%d)", current_for_printing_coil, voltage_for_printing_coil, power_for_printing_coil, nav_current_after_coil);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
            
            if (Serial.available()) 
            {
                time_t t = processSyncMessage();
                if (t != 0) 
                {
                    Teensy3Clock.set(t); // set the RTC
                    setTime(t);
                }
            }
        
            ISM303_INT.printAllSensorValues (ACC_XYZ | MAG_XYZ | TEMPERATURE);
            ISM303_A.printAllSensorValues (MAG_XYZ | MAG_MAGNITUDE | TEMPERATURE);
            ISM303_B.printAllSensorValues (MAG_XYZ | MAG_MAGNITUDE | TEMPERATURE);
            ISM303_C.printAllSensorValues (MAG_XYZ | MAG_MAGNITUDE | TEMPERATURE);
            ISM303_D.printAllSensorValues (MAG_XYZ | MAG_MAGNITUDE | TEMPERATURE);

            AccelSensors3D.X = (float)ISM303_INT.acc_f.x / ACC_100pc;
            AccelSensors3D.Y = (float)ISM303_INT.acc_f.z / ACC_100pc;
            AccelSensors3D.Z = (float)ISM303_INT.acc_f.y / ACC_100pc;
            
            MagneticSensors3D.X = -(float)ISM303_INT.mag_f.x / MAG_100pc;
            MagneticSensors3D.Y = -(float)ISM303_INT.mag_f.z / MAG_100pc;
//            MagneticSensors3D.X = (float)ISM303_INT.mag_f.x / MAG_100pc;
//            MagneticSensors3D.Y = (float)ISM303_INT.mag_f.z / MAG_100pc;
            MagneticSensors3D.Z = (float)ISM303_INT.mag_f.y / MAG_100pc;

            normalise_to_one_dot_zero_magnitude (&AccelSensors3D);
            normalise_to_one_dot_zero_magnitude (&MagneticSensors3D);


            ISM303_INT.flag_get_averages = 1;
            ISM303_A.flag_get_averages = 1;
            ISM303_B.flag_get_averages = 1;
            ISM303_C.flag_get_averages = 1;
            ISM303_D.flag_get_averages = 1;

            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_INT.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_A.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_B.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_C.txt_st);
            add_string_to_buffer_a_for_SDcard_file_writes (ISM303_D.txt_st);
            
            
            populate_TriGonioFloats_from_3Daccel_and_3Dmag (&TriGonioFloats, &AccelSensors3D, &MagneticSensors3D);
            
            float inclination = Inclination(&TriGonioFloats);
            float total_grav = TotGravFld (&TriGonioFloats);
            float total_magn = TotMagFld (&TriGonioFloats);
            float toolface_g = GravToolFace(&TriGonioFloats);
            float toolface_m = MagneticTF(&TriGonioFloats);    
            float azimuth = Azimuth(&TriGonioFloats);
            float dipangle = DipAngle(&TriGonioFloats);
            
            if (zero_rt)
            {
                MaxonMotorEncoderOnTeensy4.write(0);
                zero_rt = 0;
            }
    
            sprintf (st, " (%d)P=%d (%3.2f deg) v=%d (RPM=%3.2f)", (int)copy_encoder_nav, (int)encoder_pos, exit_shaft_orientation, (int)encoder_speed, RPM_exit_shaft);
            add_string_to_buffer_a_for_SDcard_file_writes (st);

            int valADC = adc->adc1->analogRead(A17);  

            sprintf (st, " ADC=%1.2f", valADC * 3.3 / adc->adc1->getMaxValue());
            add_string_to_buffer_a_for_SDcard_file_writes (st);
    
            sprintf (st, " i=%3.2f tg=%3.2f tm=%3.2f ", inclination, toolface_g, toolface_m);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
    
            sprintf (st, " vg=%1.4f vm=%1.4f ", total_grav, total_magn);
            add_string_to_buffer_a_for_SDcard_file_writes (st);
    
            sprintf (st, " az=%1.2f dip=%1.2f ", azimuth, dipangle);
            add_string_to_buffer_a_for_SDcard_file_writes (st);

 //           sprintf (st, " or=%d ow=%d u=%d bsr=%d bsw=%d qf=%d dbv=%d st=%d sn=%d cp=%d", I2C_master.debug_I2C_read_overflows, I2C_master.debug_I2C_write_overflows, I2C_master.buffers_in_use, I2C_master.n_bad_starts_r, 
 //                       I2C_master.n_bad_starts_w, I2C_master.buffers_quefull, I2C_master.debug_var, I2C_master.debug_n_started, I2C_master.debug_n_started_now, I2C_master.debug_n_completed);
            if (show_magnetic_extremes_ISM303_INT)
            {
                sprintf (st, " MX<%d %d %d> MY<%d %d %d> MZ<%d %d %d>", 
                            ISM303_INT.extremes.minimum.mag.x / 16, ISM303_INT.filtered_mag.x / 16, ISM303_INT.extremes.maximum.mag.x / 16, 
                            ISM303_INT.extremes.minimum.mag.y / 16, ISM303_INT.filtered_mag.y / 16, ISM303_INT.extremes.maximum.mag.y / 16, 
                            ISM303_INT.extremes.minimum.mag.z / 16, ISM303_INT.filtered_mag.z / 16, ISM303_INT.extremes.maximum.mag.z / 16);
                add_string_to_buffer_a_for_SDcard_file_writes (st);
            }
            if (show_acceleration_extremes_ISM303_INT)
            {
                sprintf (st, " GX<%d %d %d> GY<%d %d %d> GZ<%d %d %d>", 
                            ISM303_INT.extremes.minimum.acc.x / 16, ISM303_INT.filtered_acc.x / 16, ISM303_INT.extremes.maximum.acc.x / 16, 
                            ISM303_INT.extremes.minimum.acc.y / 16, ISM303_INT.filtered_acc.y / 16, ISM303_INT.extremes.maximum.acc.y / 16, 
                            ISM303_INT.extremes.minimum.acc.z / 16, ISM303_INT.filtered_acc.z / 16, ISM303_INT.extremes.maximum.acc.z / 16);
                add_string_to_buffer_a_for_SDcard_file_writes (st);
            }

            sprintf (st, "\n");
            add_string_to_buffer_a_for_SDcard_file_writes (st);
*/        
 
        }
    
        n_100ms_samples_since_boot++;
        flag_1seconds_trigger = 0;
        mutex3 = 0;
    }
    
    if (++Count_100ms_ticks_no_Alert_isr_motor > 1)   // timeout edge triggered interrupt?
    {
        SystemStatusFlags |= flag_missed_Alert_isr_motor;
        ina260_motor_alertPinAction ();    // restart after 10000 ms of no INA260 Alert pin edge interrupt
    }
    if (++Count_100ms_ticks_no_Alert_isr_coil > 2)   // timeout edge triggered interrupt?
    {
        SystemStatusFlags |= flag_missed_Alert_isr_coil;
        ina260_coil_alertPinAction ();    // restart after 10000 ms of no INA260 Alert pin edge interrupt
    }

    Count_10ms_ticks_no_100ms_isr = 0;

    SystemStatusFlags &= ~flag_in_isr_100ms;
}

time_t getTeensy3Time()
{
    return Teensy3Clock.get();
}

#define TIME_HEADER  "T"   // Header tag for serial time sync message

unsigned long processSyncMessage()
{
    unsigned long pctime = 0L;
    const unsigned long DEFAULT_TIME = 1357041600; // Jan 1 2013 

    if(Serial.find(TIME_HEADER)) 
    {
        pctime = Serial.parseInt();
        return pctime;
        if( pctime < DEFAULT_TIME) 
        { // check the value is a valid time (greater than Jan 1 2013)
          pctime = 0L; // return 0 to indicate that the time is not valid
        }
    }
    return pctime;
}

unsigned int spaceBeforeUpdate (FsFile file) 
{
  return 512 - (file.position() & 0x01FF);
}

uint32_t write_to_log_file (uint8_t *st, uint32_t k, const char *filename_first_chars, char *old_log_filename)
{
    char log_filename[40];
    TimeElements tm;
    time_t timenow;
    uint32_t s = 0;

    SystemStatusFlags |= flag_write_to_log_file;
    timenow = Teensy3Clock.get();
    breakTime(timenow, tm);
    sprintf (log_filename, "%s%02d%02d%02d%02d.txt", filename_first_chars, (1970+tm.Year) % 100, tm.Month, tm.Day, tm.Hour);

    if (strcmp(log_filename, old_log_filename) != 0)
    {
//Serial.print("sd begin ");
         if (sd.begin(SdioConfig(FIFO_SDIO))) 
        {
//Serial.println("ok");

            if (sd.exists(old_log_filename))    // make sure old file from previous hour finishes with a full text line at the bottom, so new file on next day starts with a full new line with all columns populated in that text line.
            {
                uint8_t *sst = st;
                for (uint32_t i=0; i<k; i++)
                {
                    if (*sst == '\n')
                    {
                        SystemStatusFlags |= flag_write_to_log_file_SDcard_actions;
                        {
    
                            txtFile = sd.open(old_log_filename, FILE_WRITE);
                            if (txtFile)
                            {
                                txtFile.seek(EOF);
                                txtFile.write(st, i+1);
                                txtFile.close();
                            }
                        }

                        SystemStatusFlags &= ~flag_write_to_log_file_SDcard_actions;
            
                        st = sst+1;
                        k -= (i+1);
                        break;
                    }
                    sst++;
                }
            }
            sd.end();
        }
//        else
//            Serial.println("bad");
    }
    if (k == 0)
    {
        SystemStatusFlags &= ~flag_write_to_log_file;
        return 0;
    }
    
    SystemStatusFlags |= flag_write_to_log_file_SDcard_actions;

 //Serial.print("SD BEGIN")
 
    if (sd.begin(SdioConfig(FIFO_SDIO))) 
    {
 //Serial.println(" OK");
       txtFile = sd.open(log_filename, FILE_WRITE);
        if (txtFile)
        {
            txtFile.seek(EOF);
    
            s = spaceBeforeUpdate (txtFile);    // 512 bytes blocks in SD cards, try to align to that for more efficient writing to flash
            if (s < 512)
            {
                if (k > s)
                {
                    txtFile.write (st, s); 
                    st += s; 
                    k -= s;
                }
            }
            if (k > 0)
                txtFile.write (st, k); 

            s = spaceBeforeUpdate (txtFile);
               
            strcpy (old_log_filename, log_filename);
            Count_10ms_missed_SD_write = 0;   
            SystemStatusFlags &= ~flag_write_to_SD_fails;
        }
        else
            SystemStatusFlags |= flag_write_to_SD_fails;
    
        txtFile.close();        
        sd.end();
    }
    else
    {
//        Serial.println(" BAD");

        SystemStatusFlags |= flag_write_to_SD_fails;
    }
    
    SystemStatusFlags &= ~flag_write_to_log_file_SDcard_actions;
    SystemStatusFlags &= ~flag_write_to_log_file;

    return s;
}

bool mutex2 = 0; 
int remaining_bytes_till_file_write_512_buffer_a = 512;
int remaining_bytes_till_file_write_512_buffer_b = 512;

void loop ()
{
    uint32_t i, k, copy_Buffer_used;
    uint8_t st[WRITE_TO_SD_CHUNK_SIZE+1];
    uint32_t chunk_size;
   
    SystemStatusFlags |= flag_in_loop;
 
    if (mutex2)
        SystemStatusFlags |= flag_overasking_loop_actions;
    else
    {
        mutex2 = 1;
        copy_Buffer_used = Buffer_a_used;
        chunk_size = WRITE_TO_SD_CHUNK_SIZE - 512 + remaining_bytes_till_file_write_512_buffer_a;
        if (copy_Buffer_used)
            if (copy_Buffer_used > chunk_size)
            {
                k = 0;
                for (i=0; i<chunk_size; i++)
                {   
                    st[k++] = Buffer_a[Buffer_a_read_pointer++];
                    Buffer_a_read_pointer &= BUF_SIZE_MASK;
                    Buffer_a_used--;
                }
                remaining_bytes_till_file_write_512_buffer_a = write_to_log_file (st, k, "LS_", (char*)old_log_filename_a);
            }
 
        copy_Buffer_used = Buffer_b_used;
        chunk_size = WRITE_TO_SD_CHUNK_SIZE - 512 + remaining_bytes_till_file_write_512_buffer_b;
        if (copy_Buffer_used)
            if (copy_Buffer_used > chunk_size)
            {
                k = 0;
                for (i=0; i<chunk_size; i++)
                {   
                    st[k++] = Buffer_b[Buffer_b_read_pointer++];
                    Buffer_b_read_pointer &= BUF_SIZE_MASK;
                    Buffer_b_used--;
                }
                remaining_bytes_till_file_write_512_buffer_b = write_to_log_file (st, k, "HS_", (char*)old_log_filename_b);
            }

        Count_10ms_missed_loop_cycles = 0;
        mutex2 = 0;     
    }

    SystemStatusFlags &= ~flag_in_loop;
    delay(100);

    show_interrupt_priorities ();
}
